# coding=utf-8
"""
global variables and references to global objects.
"""
from abc import ABC

# circular import error
# import UEVaultManager.tkgui.modules.DisplayContentWindowClass as DisplayContentWindow
import UEVaultManager.tkgui.modules.cls.EditCellWindowClass as EditCellWindow
import UEVaultManager.tkgui.modules.cls.EditRowWindowClass as EditRowWindow
from UEVaultManager.tkgui.modules.cls.DisplayContentWindowClass import DisplayContentWindow
from UEVaultManager.tkgui.modules.cls.FakeUEVMGuiClass import FakeUEVMGuiClass
from UEVaultManager.tkgui.modules.cls.GUISettingsClass import GUISettings
from UEVaultManager.tkgui.modules.cls.ImagePreviewWindowClass import ImagePreviewWindow
from UEVaultManager.tkgui.modules.cls.NotificationWindowClass import NotificationWindow
from UEVaultManager.tkgui.modules.cls.ProgressWindowClass import ProgressWindow
from UEVaultManager.tkgui.modules.cls.SaferDictClass import SaferDict


class WindowsRef(ABC):
    """
    Class to hold references to global windows.
    Abstractclass
    """
    uevm_gui: FakeUEVMGuiClass = None  # tkgui window , we can not use the real UEVMGuiClass because it will cause circular import error
    edit_cell: EditCellWindow = None
    edit_row: EditRowWindow = None
    display_content: DisplayContentWindow = None
    progress: ProgressWindow = None
    tool = None  # could be a ref to a ToolWindows like DBToolWindow or JsonToolWindow
    image_preview: ImagePreviewWindow = None
    notification: NotificationWindow = None

    @classmethod
    def get_properties_name(cls) -> list:
        """
        Get the propertie NAME of the class.
        :return: list of the propertie NAMES of the class.
        """
        return [prop for prop in dir(cls) if not prop.startswith('__') and not prop == '_abc_impl' and not callable(getattr(cls, prop))]

    @classmethod
    def get_properties(cls) -> list:
        """
        Get the properties of the class.
        :return: list the properties of the class.
        """
        result = []
        for name in cls.get_properties_name():
            result.append(getattr(cls, name))
        return result


# global variables that are not settings
timeout_error_count = 0  # incremented each time an image generate a request timeout
no_int_data = 0
no_float_data = 0.0
no_text_data = ''
no_bool_true_data = True
no_bool_false_data = False

# reference to the cli object of the UEVM main application (the main one, it gives all access to all the features)
# if empty, direct access to its features from this script won't be available and a message will be displayed instead
UEVM_cli_ref = None  # avoid importing classes from the UEVM main application here because it can cause circular dependencies when importing the module
#  reference to the log object of the UEVM main application.
#  If empty, log will be message printed in the console
UEVM_log_ref = None
#  reference to the default command line parser (used for help button in gui).
UEVM_parser_ref = None
# restult of the last UEVM cli command run from the GUI
UEVM_command_result = None
# global variables
s = GUISettings()  # using the shortest variable name for GUISettings for convenience
# noinspection PyTypeChecker
UEVM_cli_args: SaferDict = {}
UEVM_logger_names = []  # list of logger names used by UEVM classes
# lists of widgets that need to be enabled/disabled when some conditions change
# - add widgets to the approprriate list when they are created by calling the append_no_duplicate() method
# - the update_controls_state on window subclasses will use each list with the correct condition
stated_widgets = {
    # at least a row must be selected
    'row_is_selected': [],
    # table content has changed
    'table_has_changed': [],
    # selected item (row/page) is not the first one
    'not_first_item': [],
    # selected item (row/page)  is not the last one
    'not_last_item': [],
    # selected asset is not the first one
    'not_first_asset': [],
    # selected asset is not the last one
    'not_last_asset': [],
    # selected page is not the first one
    'not_first_page': [],
    # selected page is not the last one
    'not_last_page': [],
    # not in offline
    'not_offline': [],
    # a row is selected and the asset is owned
    'asset_is_owned': [],
    # a row is selected and the asset has an url
    'asset_has_url': [],
    # a row is selected and the asset is local
    'asset_added_mannually': [],
    # the cli object is available
    'cli_is_available': [],
    # the database mode is used
    'database_is_used': [],
    # the file mode is used
    'file_is_used': [],
    # at least a row must be selected and not in offline
    'row_is_selected_and_not_offline': [],
    # a row is selected and the asset is owned and not in offline
    'asset_is_owned_and_not_offline': [],
}


# options that can be changed in the GUI
def set_args_force_refresh(value: bool) -> None:
    """
    Set the value of the argument force_refresh. Mandadory fot the associated ttk.ckbutton to work.
    :param value: True or False.
    """
    UEVM_cli_args['force_refresh'] = value


def set_args_offline(value: bool) -> None:
    """
    Set the value of the argument offline. Mandadory fot the associated ttk.ckbutton to work.
    :param value:  True or False.
    """
    UEVM_cli_args['offline'] = value


def set_args_debug(value: bool) -> None:
    """
    Set the value of the argument debug. Mandadory fot the associated ttk.ckbutton to work.
    :param value: True or False.
    """
    UEVM_cli_args['debug'] = value


def set_args_auth_delete(value: bool) -> None:
    """
    Set the value of the argument auth_delete. Mandadory fot the associated ttk.ckbutton to work.
    :param value: True or False.
    """
    UEVM_cli_args['auth_delete'] = value


def set_args_delete_scraping_data(value: bool) -> None:
    """
    Set the value of the argument delete_scraping_data. Mandadory fot the associated ttk.ckbutton to work.
    :param value:  True or False.
    """
    UEVM_cli_args['delete_scraping_data'] = value


def set_use_threads(value: bool) -> None:
    """
    Set the value of the settings use_threads. Mandadory fot the associated ttk.ckbutton to work.
    :param value: True or False.
    """
    s.use_threads = value


def set_use_colors_for_data(value: bool) -> None:
    """
    Set the value of the settings use_colors_for_data. Mandadory fot the associated ttk.ckbutton to work.
    :param value: True or False.
    """
    s.use_colors_for_data = value


# def set_testing_switch_var(value: bool) -> None:
#   Use method UEVMGuiOptionFrame.update_gui_options() instead because it also updates the gui

# def set_debug_gui_var(value: bool) -> None:
#   Use method UEVMGuiOptionFrame.update_gui_options() instead because it also updates the gui


def set_check_asset_folders(value: bool) -> None:
    """
    Set the value of the settings check_asset_folders. Mandadory fot the associated ttk.ckbutton to work.
    :param value: True or False.
    """
    s.check_asset_folders = value


def set_browse_when_add_row(value: bool) -> None:
    """
    Set the value of the settings browse_when_add_row. Mandadory fot the associated ttk.ckbutton to work.
    :param value: True or False.
    """
    s.browse_when_add_row = value



# coding=utf-8
"""
Utilities functions and tools
These functions depend on the globals.py module and can generate circular dependencies when imported.
"""
import logging
import os
import shutil
import sys
import time
import tkinter as tk
from datetime import datetime
from io import BytesIO
from tkinter import messagebox
from typing import Optional

import requests
from PIL import Image, ImageTk
from termcolor import colored

from UEVaultManager.lfs.utils import path_join
from UEVaultManager.models.types import DateFormat
from UEVaultManager.tkgui.modules import globals as gui_g
from UEVaultManager.tkgui.modules.cls.NotificationWindowClass import NotificationWindow
from UEVaultManager.tkgui.modules.cls.ProgressWindowClass import ProgressWindow


def log_format_message(name: str, levelname: str, message: str) -> str:
    """
    Format the log message.
    :param name: name of the logger.
    :param levelname: level of log.
    :param message: message to format.
    :return: formatted message.
    """
    return f'[{name}] {levelname}: {message}'


def box_message(msg: str, level='info', show_dialog: bool = True, duration: int = -1):
    """
    Display a message box with the given message.
    :param msg: message to display.
    :param show_dialog: True to display the message in a messagebox, False to only print it on the console.
    :param duration: duration of the notification in seconds. If -1, it will use the default value set in the settings.
    :param level: level of the message (info, warning, error).
    """
    level_lower = level.lower()
    if level_lower == 'warning':
        log_warning(msg)
    elif level_lower == 'error' or level_lower == 'critical':
        log_error(msg)
        # done in log_error
        # exit(1)
    else:
        log_info(msg)
    if show_dialog:
        notification_title = ''
        if level_lower == 'debug':
            notification_title = 'Debug message' if gui_g.s.debug_mode else ''
        elif level_lower == 'warning':
            notification_title = 'Warning message'
        elif level_lower == 'error':
            notification_title = 'Error message'
        if notification_title:
            notify(title=notification_title, message=msg, duration=duration)


def box_yesno(msg: str, show_dialog: bool = True, default: bool = True) -> bool:
    """
    Display a YES/NO message box with the given message.
    :param msg: message to display .
    :param show_dialog: True to display a dialog box, False to only return default value (silent mode)
    :param default: default value to return if the dialog is not displayed.
    :return: True if the user clicked on Yes, False otherwise.
    """
    return messagebox.askyesno(title=gui_g.s.app_title, message=msg) if show_dialog else default


def box_okcancel(msg: str, show_dialog: bool = True, default: bool = True) -> bool:
    """
    Display an OK/CANCEL message box with the given message.
    :param msg: message to display.
    :param show_dialog: True to display a dialog box, False to only return  default value (silent mode)
    :param default: default value to return if the dialog is not displayed.
    :return: True if the user clicked on Yes, False otherwise.
    """
    return messagebox.askokcancel(title=gui_g.s.app_title, message=msg) if show_dialog else default


def from_cli_only_message(content='This feature is only accessible', show_dialog: bool = True) -> None:
    """
    Display a message box with a message saying that the feature is only accessible when running the application is launched using the cli.
    :param content: Optional content to add to the message.
    :param show_dialog: True to display the message in a messagebox, False to only print it on the console.
    """
    msg = f'{content} when this application is launched using the UEVM cli edit command.'
    print_msg = log_format_message(gui_g.s.app_title, 'info', colored(msg, 'yellow'))
    print(print_msg)
    if show_dialog:
        messagebox.showinfo(title=gui_g.s.app_title, message=msg)


def todo_message() -> None:
    """
    Display a message box with a message saying that the feature is not implemented yet.
    """
    msg = 'Not implemented yet'
    print_msg = log_format_message(gui_g.s.app_title, 'info', colored(msg, 'yellow'))
    print(print_msg)
    messagebox.showinfo(title=gui_g.s.app_title, message=msg)


def log_info(msg: str) -> None:
    """
    Log an info message.
    :param msg: message to log.

    Notes:
        It will use gui_g.UEVM_log_ref if defined, otherwise it will print the message on the console.
    """
    if gui_g.UEVM_log_ref is not None:
        gui_g.UEVM_log_ref.info(msg)
    else:
        print_msg = log_format_message(gui_g.s.app_title, 'info', colored(msg, 'blue'))
        print(print_msg)


def log_debug(msg: str) -> None:
    """
    Log a debug message.
    :param msg: message to log.

    Notes:
        It will use gui_g.UEVM_log_ref if defined, otherwise it will print the message on the console.
        This message will only be logged if the debug mode is enabled.
    """
    if not gui_g.s.debug_mode:
        return
    if gui_g.UEVM_log_ref is not None:
        # ensure that the debug messages will be logged even if the log level not set to DEBUG in the cli
        log_level = gui_g.UEVM_log_ref.level
        if log_level == logging.DEBUG:
            gui_g.UEVM_log_ref.debug(msg)
        else:
            gui_g.UEVM_log_ref.info(msg)
    else:
        print_msg = log_format_message(gui_g.s.app_title, 'Debug', colored(msg, 'light_grey'))
        print(print_msg)


def log_warning(msg: str) -> None:
    """
    Log a warning message.
    :param msg: message to log.

    Notes:
        It will use gui_g.UEVM_log_ref if defined, otherwise it will print the message on the console.
    """
    if gui_g.UEVM_log_ref is not None:
        gui_g.UEVM_log_ref.info(msg)
    else:
        print_msg = log_format_message(gui_g.s.app_title, 'Warning', colored(msg, 'magenta'))
        print(print_msg)


def log_error(msg: str) -> None:
    """
    Log an error message.
    :param msg: message to log.

    Notes:
        The application will (normally) exit after logging the message. Sometimes it doesn't work (check ?)
        It will use gui_g.UEVM_log_ref if defined, otherwise it will print the message on the console.
    """
    if gui_g.UEVM_log_ref is not None:
        gui_g.UEVM_log_ref.error(msg)
    else:
        print_msg = log_format_message(gui_g.s.app_title, 'Error', msg)
        print(print_msg)
        exit_and_clean_windows()


def resize_and_show_image(image: Image, canvas: tk.Canvas, scale: float = 1.0, x: int = -1, y: int = -1) -> None:
    """
    Resize the given image and display it in the given canvas.
    :param image: image to display.
    :param canvas: canvas to display the image in.
    :param scale: scale to apply to the image.
    :param x: x coordinate of the image. If -1, the image will be centered.
    :param y: y coordinate of the image. If -1, the image will be centered.
    """
    # Resize the image while keeping the aspect ratio
    target_height = int(gui_g.s.preview_max_height * scale)
    aspect_ratio = float(image.width * scale) / float(image.height * scale)
    target_width = int(target_height * aspect_ratio)
    resized_image = image.resize((target_width, target_height), Image.Resampling.BILINEAR)
    tk_image = ImageTk.PhotoImage(resized_image)
    anchor = tk.NW if x == -1 and y == -1 else tk.CENTER
    # Calculate center coordinates
    x = max(0, (canvas.winfo_width() - tk_image.width()) // 2) if x == -1 else x
    y = max(0, (canvas.winfo_height() - tk_image.height()) // 2) if y == -1 else y
    canvas.create_image(x, y, anchor=anchor, image=tk_image)
    canvas.image = tk_image


def show_asset_image(image_url: str, canvas_image=None, scale: float = 1.0, x: int = -1, y: int = -1, timeout=(4, 4)) -> bool:
    """
    Show the image of the given asset in the given canvas.
    :param image_url: url of the image to display.
    :param canvas_image: canvas to display the image in.
    :param scale: scale to apply to the image.
    :param x: x coordinate of the image. If -1, the image will be centered.
    :param y: y coordinate of the image. If -1, the image will be centered.
    :param timeout: timeout for the request. Could be a float or a tuple of float (connect timeout, read timeout).
    :return: True if the image has been displayed, False otherwise.
    """
    if gui_g.s.offline_mode:
        # could be usefull if connexion is slow
        show_default_image(canvas_image)
        return False
    if canvas_image is None or not image_url or str(image_url) in gui_g.s.cell_is_empty_list:
        return False
    try:
        # print(image_url)
        # noinspection DuplicatedCode
        if not os.path.isdir(gui_g.s.asset_images_folder):
            os.mkdir(gui_g.s.asset_images_folder)
        image_filename = path_join(gui_g.s.asset_images_folder, os.path.basename(image_url))
        # Check if the image is already cached
        if os.path.isfile(image_filename) and (time.time() - os.path.getmtime(image_filename)) < gui_g.s.image_cache_max_time:
            # Load the image from the cache folder
            image = Image.open(image_filename)
        else:
            response = requests.get(image_url, timeout=timeout)
            image = Image.open(BytesIO(response.content))
            with open(image_filename, "wb") as file:
                file.write(response.content)
        resize_and_show_image(image=image, canvas=canvas_image, scale=scale, x=x, y=y)
        return True
    except Exception as error:
        log_warning(f'Error showing image: {error!r}')
        gui_g.timeout_error_count += 1
        # check error timeout
        if gui_g.timeout_error_count >= 5:
            box_message(
                f'The application had {gui_g.timeout_error_count} timeout errors when loading images.\nIt is going offline to avoid been too slow.\nTo fix that, check you internet connection.\nYou can disabled offline mode in the "Show options" panel, or by restarting the application.',
                level='warning'
            )
            gui_g.timeout_error_count = 0
            gui_g.s.offline_mode = True
            return False


def show_default_image(canvas_image=None) -> None:
    """
    Show the default image in the given canvas.
    :param canvas_image: canvas to display the image in.
    """
    if canvas_image is None:
        return
    try:
        # Load the default image
        if os.path.isfile(gui_g.s.default_image_filename):
            def_image = Image.open(gui_g.s.default_image_filename)
            resize_and_show_image(def_image, canvas_image)
    except Exception as error:
        log_warning(f'Error showing default image {gui_g.s.default_image_filename} cwd:{os.getcwd()}: {error!r}')


def json_print_key_val(json_obj, indent=4, print_result=True, output_on_gui=False) -> None:
    """
    Pretty prints a JSON object in a simple 'key: value' format.
    :param json_obj:  The JSON object to print.
    :param indent: number of spaces to indent each level.
    :param print_result: determines whether to print the result.
    :param output_on_gui: determines whether to print the result on the GUI.
    :return: pretty printed JSON object.
    """

    def _process(obj, level=0):
        lines = []
        indent_str = ' ' * indent * level

        if isinstance(obj, dict):
            for key, value in obj.items():
                if isinstance(value, (dict, list)):
                    lines.append(f'{indent_str}{key}:')
                    lines.extend(_process(value, level + 1))
                else:
                    lines.append(f'{indent_str}{key}: {value}')
        elif isinstance(obj, list):
            for item in obj:
                lines.extend(_process(item, level))
        else:
            lines.append(f'{indent_str}{obj}')

        return lines

    result = '\n'.join(_process(json_obj))

    if print_result:
        if output_on_gui and gui_g.WindowsRef.display_content is not None:
            gui_g.WindowsRef.display_content.display(result)
        else:
            print(result)


def custom_print(text='', keep_mode=True) -> None:
    """
    Print the given text on the GUI if it's available, otherwise print it on the console².
    :param text: text to print.
    :param keep_mode: whether to keep the existing content when adding a new one.
    """
    if gui_g.WindowsRef.display_content is not None:
        gui_g.WindowsRef.display_content.display(content=text, keep_mode=keep_mode)
    else:
        print(text)


def get_tk_root(container) -> Optional[tk.Tk]:
    """
    Get the root window.
    :param container:  the container window or object.
    :return: root window.
    """
    if container is None:
        return None
    root = None
    # get the root window to avoid creating multiple progress windows
    try:
        # a tk window child class
        root = container.winfo_toplevel()
    except (AttributeError, tk.TclError):
        # an editableTable class
        try:
            root = container.get_container.winfo_toplevel()
        except (AttributeError, tk.TclError):
            pass
    return root


def show_progress(
    parent=None,
    text='Working...Please wait',
    width=500,
    height=120,
    max_value_l=0,
    show_progress_l=False,
    show_btn_stop_l=False,
    quit_on_close: bool = False,
    keep_existing: bool = False,
    function: callable = None,
    function_parameters: dict = None,
    force_new_window: bool = False
) -> Optional[ProgressWindow]:
    """
    Show the progress window. If the progress window does not exist, it will be created.
    :param parent: parent window. Could be None.
    :param text: text to display in the progress window.
    :param width: width of the progress window.
    :param height: height of the progress window.
    :param max_value_l: maximum value of the progress bar.
    :param show_progress_l: whether to show the progress bar.
    :param show_btn_stop_l: whether to show the stop button.
    :param quit_on_close: whether to quit the application when the window is closed.
    :param keep_existing: whether to keep the existing content when adding a new one.
    :param function: function to execute.
    :param function_parameters: parameters of the function.
    :param force_new_window: whether to force the creation of a new progress window.
    :return: progress window.
    It will create a new progress window if one does not exist and update parent._progress_window
    """
    pw = None
    root = get_tk_root(parent)
    create_a_new_progress_window = force_new_window
    if not root:
        return None
    try:
        # check if a progress window already exists
        pw = root.progress_window
        if force_new_window:
            pw.close_window(True)
        else:
            pw.get_text()  # a call to test if the window is still active
    except (tk.TclError, AttributeError, RuntimeError):
        create_a_new_progress_window = True

    if create_a_new_progress_window:
        pw = ProgressWindow(
            title=gui_g.s.app_title,
            parent=parent,
            icon=gui_g.s.app_icon_filename,
            width=width,
            height=height,
            show_btn_stop=show_btn_stop_l,
            show_progress=show_progress_l,
            max_value=max_value_l,
            quit_on_close=quit_on_close,
            function=function,
            function_parameters=function_parameters
        )
        root.progress_window = pw
    try:
        # here the pw could have already been close if it was modal or using thhreads
        pw.set_value(0)  # reset the progress bar, needed if the progress window is reused
        pw.set_activation(False)
        if keep_existing:
            text = pw.get_text() + '\n' + text
        pw.set_text(text)
    except (tk.TclError, AttributeError) as error:
        log_debug(f'Error showing progress window: {error!r}')
        pw = None
        root.progress_window = None
        gui_g.progress_window_ref = None
    return pw


def close_progress(parent) -> None:
    """
    Close the progress window.
    :param parent: parent window.
    It accesses to the parent.progress_window property
    """
    root = get_tk_root(parent)
    if root and root.progress_window is not None:
        root.progress_window.close_window()
        root.progress_window = None


def create_file_backup(file_src: str, logger: logging.Logger = None, backups_folder: str = '', backup_to_keep: int = -1, suffix: str = '') -> str:
    """
    Create a backup of a file.
    :param file_src: path to the file to back up.
    :param logger: logger to use to display info. Could be None.
    :param backups_folder: path to the backup folder.
    :param backup_to_keep: number of backup to keep. Set to 0 to keep all the backups or to -1 to use the default value in the settings.
    :param suffix: suffix to add to the backup file name. If None, it will use the current date.
    :return: full path to the backup file.

    Notes:
        If backups_folder to None, the backup will be done in the same folder as the source file,
        If backups_folder to '', the backup will be done in the default backup folder set in the config file.
    """
    file_backup = ''
    if not file_src or not os.path.isfile(file_src):
        return ''
    file_name_no_ext, file_ext = os.path.splitext(file_src)
    suffix = suffix if suffix else f'{datetime.now().strftime(DateFormat.file_suffix)}'
    try:
        file_name_no_ext = os.path.basename(file_name_no_ext)
        file_backup = f'{file_name_no_ext}_{suffix}{file_ext}{gui_g.s.backup_file_ext}'
        if backups_folder is None:
            backups_folder = os.path.dirname(file_src)
        elif backups_folder == '':
            backups_folder = gui_g.s.backups_folder
        if not os.path.isdir(backups_folder):
            os.mkdir(backups_folder)
        file_backup = path_join(backups_folder, file_backup)
        shutil.copy(file_src, file_backup)
        if logger is not None:
            logger.info(f'File {file_src} has been copied to {file_backup}')
    except (Exception, ):
        if logger is not None:
            logger.info(f'File {file_src} coulnd not been backed up')
            return ''
    backup_to_keep = gui_g.s.backup_files_to_keep if backup_to_keep == -1 else backup_to_keep
    if backup_to_keep > 0:
        # delete old backups
        backup_list = []
        file_name_no_ext = os.path.basename(file_name_no_ext)
        for file in os.listdir(backups_folder):
            filename = os.path.basename(file)
            if filename.startswith(file_name_no_ext) and filename.endswith(gui_g.s.backup_file_ext):
                backup_list.append(file)
        backup_list.sort()
        if len(backup_list) > backup_to_keep:
            for file in backup_list[:-backup_to_keep]:
                file_to_delete = path_join(backups_folder, file)
                os.remove(file_to_delete)
                if logger is not None:
                    logger.info(f'Backup File {file_to_delete} has been deleted')
    return file_backup


def update_loggers_level(logger: logging.Logger = None, debug_value=None) -> None:
    """
    Change the logger level of debug depending on the debug mode.
    :param logger: logger.
    :param debug_value: value to set. If None, it will use the value of gui_g.s.debug_mode.

    Notes:
        Will also update all the loggers level of the UEVM classes.
        Call this function when the debug mode is changed.
    """
    if logger is not None:
        if logger.name not in gui_g.UEVM_logger_names:
            gui_g.UEVM_logger_names.append(logger.name)
    debug_value = gui_g.s.debug_mode if debug_value is None else debug_value
    for logger_name in gui_g.UEVM_logger_names:
        logger = logging.getLogger(logger_name)
        logger.setLevel(level=logging.DEBUG if debug_value else logging.INFO)


def make_modal(window: tk.Toplevel = None, wait_for_close=True, widget_to_focus: tk.Widget = None) -> None:
    """
    Make the given window modal.
    :param window: window to make modal.
    :param wait_for_close: whether to wait for the window to be closed before continuing.
    :param widget_to_focus: widget to focus when the window is opened.
    """
    window.grab_set()
    if widget_to_focus is not None:
        widget_to_focus.focus_set()
    else:
        window.focus_set()
    if wait_for_close:
        # Note: this will block the main window
        window.wait_window()


def set_widget_state(widget: tk.Widget, is_enabled: bool, text_swap: {} = None) -> None:
    """
    Enable or disable a widget.
    :param widget: widget to update.
    :param is_enabled: whether to enable the widget, if False, disable it.
    :param text_swap: dict {'normal':text, 'disabled':text} to swap the text of the widget depending on its state.
    """
    if widget is not None:
        state = tk.NORMAL if is_enabled else tk.DISABLED
        state_inversed = tk.NORMAL if not is_enabled else tk.DISABLED
        try:
            # noinspection PyUnresolvedReferences
            widget.config(state=state)
            current_text = widget.cget('text')
            if text_swap is not None and text_swap.get(state, '') == current_text:
                swapped_text = text_swap.get(state_inversed, '')
                if swapped_text:
                    # noinspection PyUnresolvedReferences
                    widget.config(text=swapped_text)
        except tk.TclError:
            pass


def enable_widget(widget) -> None:
    """
    Enable a widget.
    :param widget: widget to update.
    """
    set_widget_state(widget, True)


def disable_widget(widget) -> None:
    """
    Disable a widget.
    :param widget: widget to update.
    """
    set_widget_state(widget, False)


def set_widget_state_in_list(list_of_widget: [], is_enabled: bool, text_swap: {} = None) -> None:
    """
    Enable or disable a widget.
     :param list_of_widget: list of widgets to update.
    :param is_enabled: whether to enable the widget, if False, disable it.
    :param text_swap: dict {'normal':text, 'disabled':text} to swap the text of the widget depending on its state.
    """
    for widget in list_of_widget:
        set_widget_state(widget, is_enabled, text_swap)


def enable_widgets_in_list(list_of_widget: []) -> None:
    """
    Enable a list of widgets.
    :param list_of_widget: list of widgets to enable.
    """
    for widget in list_of_widget:
        enable_widget(widget)


def disable_widgets_in_list(list_of_widget: []) -> None:
    """
    Disable a list of widgets.
    :param list_of_widget: list of widgets to disable.
    :return:
    """
    for widget in list_of_widget:
        disable_widget(widget)


def update_widgets_in_list(is_enabled: bool, list_name: str, text_swap=None) -> None:
    """
    Update the state of a list of widgets.
    :param is_enabled: True to enable the widgets, False to disable them.
    :param list_name: name of the list of widgets to update.
    :param text_swap: dict {'normal':text, 'disabled':text} to swap the text of the widget depending on its state.
    """
    widget_list = gui_g.stated_widgets.get(list_name, [])
    set_widget_state_in_list(widget_list, is_enabled=is_enabled, text_swap=text_swap)


def exit_and_clean_windows(code: int = 1):
    """
    Exit the application and clean all the windows before.
    :return:
    """
    window_list = gui_g.WindowsRef.get_properties()
    for window in window_list:
        if window is not None:
            window.quit()
            window.destroy()
    sys.exit(code)


def parse_callable(callable_string: str) -> (str, dict):
    """
    Parse a callable string and return the function name and parameters.
    :param callable_string:
    :return: (function name, parameters)
    """
    try:
        func_prefix = 'filter_'  # prefix to add to the function name if not already present
        func_parts = callable_string.split('##')
        func_name = func_parts[0]
        func_name = func_prefix + func_name if not func_name.startswith(func_prefix) else func_name
        func_params = func_parts[1:]
        return func_name, func_params
    except (Exception, ):
        return '', []


def save_image_to_png(image: tk.PhotoImage, filename: str) -> bool:
    """
    Save an image to a PNG file.
    :param image: image to save.
    :param filename: filename to save the image to. The extension will be changed to .png if needed.
    :return: True if the image was saved, False otherwise.
    """
    try:
        filename = os.path.normpath(filename)
        filename = os.path.splitext(filename)[0] + '.png'
        img_pil = ImageTk.getimage(image)
        size = img_pil.size  # get the size of the image
        if img_pil.mode in ('RGBA', 'LA'):
            background = Image.new(img_pil.mode[:-1], size, '#000')
            # noinspection PyTypeChecker
            background.paste(img_pil, img_pil.split()[-1])
            img_pil = background
        img_pil.save(filename, format='PNG', subsampling=0, quality=100)
        img_pil.close()
        return True
    except (Exception, ):
        return False


def notify(message: str = '', title: str = '', duration: int = -1) -> Optional[NotificationWindow]:
    """
    Display a notification message.
    :param message: message to display.
    :param title: title of the notification.
    :param duration: duration of the notification in seconds. If -1, it will use the default value set in the settings.
    """
    if not message:
        return None
    if duration == -1:
        duration = gui_g.s.notification_time
    nw = NotificationWindow(title=title or gui_g.s.app_title, message=message, duration=duration)
    nw.show()
    return nw


def close_notify():
    """
    Close the notiification Widows
    """
    if gui_g.WindowsRef:
        gui_g.WindowsRef.notification.close()


def copy_widget_value_to_clipboard(container, event) -> bool:
    """
    Copy the value of a widget to the clipboard.
    :param container: container of the widget.
    :param event: event that triggered the function.
    :return: True if the value was copied, False otherwise.
    """
    try:
        widget = event.widget
        if widget.widgetName != 'ttk::frame':
            # get the widget content depending on the widget type
            if widget.widgetName in ['ttk::combobox', 'ttk::spinbox', 'ttk::entry']:
                value = widget.get()
            elif widget.widgetName == 'ttk::checkbutton':
                value = widget.switch_state(event=event)
            elif widget.widgetName == 'tk.text':
                value = widget.get('1.0', tk.END)
            elif 'HTMLScrolledText' in str(type(widget)):  # HTMLScrolledText
                value = widget.get('1.0', tk.END)
            else:  # any ExtendedWidget
                value = widget.get_content()
            # copy the value in clipboard
            container.clipboard_clear()
            container.clipboard_append(value)
            notify(f'Widget content has been copied into clipboard', duration=3000)
            return True
    except AttributeError:
        return False



# coding=utf-8
"""
Utilities functions and tools
These functions DO NOT depend on the globals.py module and be freely imported.
"""
import ast
import ctypes as ct
import datetime
import os
import subprocess
import sys
import uuid
from typing import Optional

import ttkbootstrap as ttk
from screeninfo import get_monitors
from ttkbootstrap.publisher import Publisher

from UEVaultManager.lfs.utils import path_join


def log(message: str) -> None:
    """
    Log a message in the console.
    :param message: message to log.
    """
    # keep this function as simple as possible to avoid circular import
    print(message)


def path_from_relative_to_absolute(path: str) -> str:
    """
    Build the path of the file to reference relative to the currently running script.
    :param path: relative path to the file. If the path is already absolute, it is returned as is.
    :return: absolute path of the file.
    """
    if os.path.isabs(path):
        return os.path.normpath(path)

    try:
        # PyInstaller creates a temp folder and stores path in _MEIPASS
        # noinspection PyProtectedMember,PyUnresolvedReferences
        # base_path = sys._MEIPASS
        current_script_directory = sys._MEIPASS
        # when launched from pyinstaller, the path has been copied at the root of the folder where the script is run. So we nne to remove all the relative path indicator
        path = os.path.basename(path)
    except AttributeError:
        # base_path = os.path.abspath(".")
        current_script_path = os.path.abspath(__file__)
        current_script_directory = os.path.dirname(current_script_path)

    absolute_path = path_join(current_script_directory, path)
    return os.path.normpath(absolute_path)


def center_window_on_screen(screen_index: int, width: int, height: int, set_size: bool = True) -> str:
    """
    Calculate the geometry of the window to display in the center of the given screen.
    :param screen_index: index of the screen to use.
    :param width: width of the window.
    :param height: height of the window.
    :param set_size: wether to set the size of the window or not.
    :return: geometry string to use to display the window in the center of the screen.
    """
    x, y = get_center_screen_positions(screen_index, width, height)
    x = f'+{x}'  # keep the sign !
    y = f'+{y}'  # keep the sign !
    geometry: str = f'{width}x{height}{x}{y}' if set_size else f'{x}{y}'
    return geometry


def get_screen_positions(screen_index: int) -> (int, int, int, int):
    """
    Return the x and y positions of the given screen.
    :param screen_index: index of the screen to use.
    :return: (x, y, w, h) positions of the given screen.
    """
    monitors = get_monitors()
    if screen_index > len(monitors):
        log(f'The screen #{screen_index} is not available. Using 0 as screen index.')  # no use of log functions here to prevent circular import
        screen_index = 0
    target_screen = monitors[screen_index]
    return target_screen.x, target_screen.y, target_screen.width, target_screen.height


def get_center_screen_positions(screen_index: int, width: int, height: int) -> (int, int):
    """
    Return the x and y positions of the window to display in the center of the given screen.
    :param screen_index: index of the screen to use.
    :param width: width of the window.
    :param height: height of the window.
    :return: (x, y) positions of the window to display in the center of the given screen.
    """
    screen_x, screen_y, screen_w, screen_h = get_screen_positions(screen_index)
    x = screen_x + (screen_w - width) // 2
    y = screen_y + (screen_h - height) // 2
    return x, y


def set_custom_style(theme_name='lumen', font=('Arial', 10, 'normal')):
    """
    Set the custom style for the application.
    :return: Style object.
    """
    try:
        style = ttk.Style(theme_name)
    except (Exception, ):
        # free the subscriber widgets to the Style. As it, it could be used by other window without issue
        Publisher.clear_subscribers()
        style = ttk.Style(theme_name)

    # option possible for ttk widgets:
    # TButton, TCheckbutton, TCombobox, TEntry, TFrame, TLabel, TLabelFrame, TMenubutton, TNotebook, TProgressbar, TRadiobutton,
    # TScale, TScrollbar, TSeparator, TSizegrip, Treeview, TPanedwindow,
    # Horizontal.TProgressbar or Vertical.TProgressbar (depending on the orient option),
    # Horizontal.TScale or Vertical.TScale (depending on the orient option),
    # Horizontal.TScrollbar or Vertical.TScrollbar (depending on the orient option)
    style.configure('TLabel', font=font, spacing=1, padding=2)
    style.configure('TButton', font=font, spacing=1, padding=2)
    style.configure('TEntry', font=font, spacing=1, padding=2)
    style.configure('TFrame', font=font, spacing=1, padding=1)
    style.configure('TCombobox', font=font, spacing=1, padding=1)
    style.configure('TLabelFrame', font=font, spacing=1, padding=1)
    return style


def set_toolbar_style(tk_window) -> None:
    """
    Remove the minimize and maximize buttons from a tkinter window.
    This version is compatible with Windows AND Non-windows OS
    # see https://stackoverflow.com/questions/2969870/removing-minimize-maximize-buttons-in-tkinter.
    :param tk_window: tkinter window.
    """
    try:
        set_window_pos = ct.windll.user32.SetWindowPos
        set_window_long = ct.windll.user32.SetWindowLongPtrW
        get_window_long = ct.windll.user32.GetWindowLongPtrW
        get_parent = ct.windll.user32.GetParent
    except AttributeError:
        # Non-windows OS
        log('Non-windows OS detected. No need to remove the minimize and maximize buttons from the window.')
        return
    # Identifiers
    gwl_style = -16
    ws_minimizebox = 131072
    ws_maximizebox = 65536
    swp_nozorder = 4
    swp_nomove = 2
    swp_nosize = 1
    swp_framechanged = 32
    hwnd = get_parent(tk_window.winfo_id())
    old_style = get_window_long(hwnd, gwl_style)  # Get the style
    new_style = old_style & ~ws_maximizebox & ~ws_minimizebox  # New style, without max/min buttons
    set_window_long(hwnd, gwl_style, new_style)  # Apply the new style
    set_window_pos(hwnd, 0, 0, 0, 0, 0, swp_nomove | swp_nosize | swp_nozorder | swp_framechanged)  # Updates


def set_icon_and_minmax(tk_window, icon=None) -> None:
    """
    Set the icon and remove the min/max buttons of the window if no icon is provided.
    :param tk_window: tkinter window.
    :param icon: path of the icon to use.
    """
    if icon is None:
        # remove the min/max buttons of the window
        # this code works on Window only
        # tk_window.attributes('-toolwindow', True)
        tk_window.after(300, lambda: set_toolbar_style(tk_window))
    else:
        # windows only (remove the minimize/maximize buttons and the icon)
        icon = path_from_relative_to_absolute(icon)
        if icon and os.path.isfile(icon):
            try:
                tk_window.iconbitmap(icon)
            except Exception as error:
                # in linux, the ico can exist but not be readable
                log(f'Error while setting the icon: {error!r}')


def create_empty_file(file_path: str) -> (bool, str):
    """
    Create an empty file.
    :param file_path: path of the file to create.
    :return: (True if path was valid, the corrected path of the file).
    """
    path, file = os.path.split(file_path)
    is_valid, path = check_and_get_folder(path)
    file_path = path_join(path, file)
    open(file_path, 'w').close()
    return is_valid, file_path


def check_and_get_folder(folder_path: str) -> (bool, str):
    """
    Check if the folder exists. If not, create it or use the default one.
    :param folder_path: path of the folder to check.
    :return: (True if path was valid, the corrected path of the folder).
    """
    path = folder_path
    is_valid = True
    if not os.path.exists(path):
        try:
            os.makedirs(path)
        except (OSError, PermissionError) as error:
            is_valid = False
            log(f'Error while creating the directory {path}: {error!r}')
            if home_dir := os.environ.get('XDG_CONFIG_HOME'):
                path = path_join(home_dir, 'UEVaultManager')
            else:
                path = os.path.expanduser('~/.config/UEVaultManager')
            if not os.path.exists(path):
                os.makedirs(path)
            path = os.path.normpath(path)
            log(f'The following folder {path} will be used as default')
    return is_valid, path


def convert_to_bool(value) -> bool:
    """
    Convert a value to a boolean. Useful for None values.
    :param value: value to convert. If the value is not a boolean, it will be converted to a string and then to a boolean.
    :return: boolean value.
    """
    try:
        if str(value).lower() in ('1', '1.0', 'true', 'yes', 'y', 't'):
            return True
        else:
            return False
    except (TypeError, ValueError):
        return False


def convert_to_int(value, prefix: str = '') -> int:
    """
    Convert a value to an integer.
    :param value: value to convert.
    :param prefix: prefix to remove from the value before converting it to an integer.
    :return: integer value or 0 if the value is None or not an integer.
    """
    # remove prefix if any
    if prefix and isinstance(value, str) and value.startswith(prefix):
        value = value[len(prefix):]
    try:
        value = int(value)
        return value
    except (TypeError, ValueError):
        return 0


def convert_to_float(value) -> float:
    """
    Convert a value to a float. Useful for None values.
    :param value: value to convert.
    :return: float value or 0.0 if the value is None or not a float.
    """
    try:
        value = float(value)
        return value
    except (TypeError, ValueError):
        return 0.0


def convert_to_datetime(value: str, formats_to_use='%Y-%m-%d %H:%M:%S', default=None) -> datetime.datetime:
    """
    Convert a value to a datetime object.
    :param value: value to convert. If the value is not a datetime, it will be converted to a string and then to a datetime.
    :param formats_to_use: list of format to use to trye to convert the value. They will be tried in order.
    :param default: default value to return if the conversion fails. If None, the default value is 1970-01-01 00:00:00.
    :return: datetime value.
    """
    if default is None:
        default = datetime.datetime(1970, 1, 1)
    if not value:
        return default
    if not isinstance(formats_to_use, list):
        formats_to_use = [formats_to_use]
    for date_format in formats_to_use:
        try:
            value_date = datetime.datetime.strptime(value, date_format)
            if value_date != datetime.datetime(1970, 1, 1):
                return value_date
        except (TypeError, ValueError, AttributeError):
            continue

    return default


def convert_to_str_datetime(value, date_format='%Y-%m-%d %H:%M:%S', default=None) -> str:
    """
    Convert a value to a datetime string.
    :param value: value to convert. If the value is not a datetime.
    :param date_format: format of value.
    :param default: default value to return if the conversion fails. If None, the default value is 1970-01-01 00:00:00.
    :return: string value of the datetime.
    """
    # we convert only datetime object
    if isinstance(value, str):
        return value

    if default is None:
        default = datetime.datetime(1970, 1, 1).strftime(date_format)
    if not value:
        return default
    try:
        return value.strftime(date_format)
    except (TypeError, ValueError, AttributeError):
        return default


def is_an_int(value, prefix: str = '', prefix_is_mandatory: bool = False) -> bool:
    """
    Check if a value is an integer.
    :param value: value to check.
    :param prefix: prefix to remove from the value before checking if it is an integer.
    :param prefix_is_mandatory: wether the prefix is mandatory or not.
    :return: True if the value is an integer, False otherwise.
    """
    # check prefix if any
    if prefix and isinstance(value, str) and value.startswith(prefix):
        value = value[len(prefix):]
    elif prefix_is_mandatory:
        return False
    try:
        float_n = float(value)
        int_n = int(float_n)
    except (Exception, ):
        return False
    else:
        return float_n == int_n


def create_uid() -> str:
    """
    Create a unique id.
    :return: unique id.
    """
    return str(uuid.uuid4())[:8]


# def create_id_from_origin(string: str) -> str:
#     """
#     Create a hash from a string.
#     :return: unique hash.
#     """
#     # hash may return different values for same string, as PYTHONHASHSEED Value will change everytime you run your program. You may want to set it to some fixed value. Read here
#     # ignoring 1st character as it may be the negative sign.
#     return str(hash(string))[1:13]


def shorten_text(url: str, limit: int = 30, prefix: str = '...') -> str:
    """
    Shorten an url. Get its last part
    :param url:  the url to shorten.
    :param limit: limit of characters to keep.
    :param prefix: prefix to add to the shorted url (if it has been shorted).
    :return: shortened url.
    """
    if len(url) < limit:
        return url
    else:
        return prefix + url[-limit:]


def extract_variables_from_url(url: str) -> dict:
    """
    Extract variables from an url.
    :param url: url to extract variables from.
    :return: dict containing the variable.
    """
    result = {}
    url_parts = url.split('?')
    if len(url_parts) == 2:
        url_params = url_parts[1].split('&')
        extracted_data = {}
        for param in url_params:
            key, value = param.split('=')
            extracted_data[key] = value
        result = extracted_data
    return result


def open_folder_in_file_explorer(folder_path) -> bool:
    """
    Open a folder in the file explorer.
    :param folder_path: path of the folder to open.
    """
    if os.path.exists(folder_path):
        try:
            # For Windows
            if os.name == 'nt':
                # if we use check = True, it will raise an error even if the folder is opened, because this command always return 1 on windows
                process = subprocess.run(['explorer', folder_path], check=False)
                result = process.returncode == 0 or process.returncode == 1
            # For macOS
            elif os.name == 'posix':
                process = subprocess.run(['open', folder_path], check=False)
                result = process.returncode == 0
            # For Linux
            else:
                process = subprocess.run(['xdg-open', folder_path], check=False)
                result = process.returncode == 0
            return result
        except subprocess.CalledProcessError as error:
            log(f'Failed to open {folder_path} in file explorer.Error code: {error!r}')
            return False
    else:
        return False


def append_no_duplicate(list_to_append: list, items: any, ok_if_exists: object = False) -> bool:
    """
    Append some items value to a list. Could raise an error if an object is already in the list.
    :param list_to_append: list to append to.
    :param items: items to append to the list. Could be a single object or a list of items.
    :param ok_if_exists: wether no error will be raised if an object is already in the list.
    :return: True if all the items were appended, False if at least one was already in the list.
    """
    if not isinstance(items, list):
        items = [items]
    for item in items:
        if item not in list_to_append:
            list_to_append.append(item)
        else:
            if not ok_if_exists:
                raise ValueError(f'append_no_duplicate method: Value {item} already in list {list_to_append}')
            return False
    return True


def merge_lists_or_strings(list_to_merge, list_to_append) -> list:
    """
    Merge 2 lists (or strings) without duplicates.
    :param list_to_merge: list to merge. Could be a list or a string of values separated by commas.
    :param list_to_append: list to append Could be a list or a string of values separated by commas.
    :return: merged list.
    """
    if isinstance(list_to_merge, str):
        list_to_merge = list_to_merge.split(',') if list_to_merge else []
    if isinstance(list_to_append, str):
        list_to_append = list_to_append.split(',') if list_to_append else []
    # merge the 2 lists without duplicates
    # old method
    # for folder in set(list_to_merge + list_to_append):
    #     if folder not in list_to_merge:
    #         list_to_merge.append(folder)
    # return list_to_merge
    # shorter method
    return list(set(list_to_merge + list_to_append))  # no sorting here because the order could stay first In first Out


def remove_last_suffix(string: str, separator: str = '_') -> str:
    """
    Remove the last suffix from a string.
    :param string: String to remove the suffix from.
    :param separator: separator to use to split the string.
    :return: string without the last suffix.
    """
    parts = string.split(separator)
    if len(parts) > 1:
        parts.pop()
        return separator.join(parts)
    return string


def format_size(size: int, precision: int = 1) -> str:
    """
    Format a size in bytes to a human-readable string.
    :param size: size to format.
    :param precision: number of digits after the decimal point.
    :return: formatted size.
    """
    suffixes = ['B', 'KB', 'MB', 'GB', 'TB']
    suffix_index = 0
    while size > 1024 and suffix_index < 4:
        suffix_index += 1  # increment the index of the suffix
        size /= 1024.0  # apply the division
    return f'{size:.{precision}f}{suffixes[suffix_index]}'


def check_and_convert_list_to_str(str_or_list) -> str:
    """
    Check if the given parameter is a list and convert it to a string, else return the given parameter.
    :param str_or_list: string or list to convert.
    :return: converted string or the given parameter.
    """
    result = str_or_list
    # Note: can not use isinstance because 'Index', 'ndarray' and 'dict_values' are not recognized as types
    if type(str_or_list).__name__ in ('list', 'dict_values', 'Index', 'ndarray'):
        result = ','.join([str(value) for value in str_or_list])
    return result


def get_and_check_release_info(data_to_check, empty_values: list = None) -> Optional[list]:
    """
    Check if the given data is a list of release info.
    :param data_to_check: data to check.
    :param empty_values: list of values to consider as empty.
    :return: list of release info or None if the data is not valid.
    """
    if not data_to_check or (empty_values and data_to_check in empty_values):
        return []
    try:
        max_tries = 3
        tries = 0
        # Note:
        #   here we use ast.literal_eval instead of json.loads because it can raise an error if the string came from a datatable and uses ' instead of " for string literals
        while '[{\"id\":' in str(data_to_check) and tries < max_tries:
            # fix a multiple encoding issue. Should not occur
            # data_to_check = json.loads(data_to_check)
            tries += 1
            data_to_check = ast.literal_eval(data_to_check)
        if isinstance(data_to_check, str):
            # data_to_check = json.loads(data_to_check)
            data_to_check = ast.literal_eval(data_to_check)
        if not isinstance(data_to_check, list):
            return None
        return list(data_to_check)
    except (Exception, ):
        return None



# coding=utf-8
"""
Implementation for:
- UCResponse: Response object for the undetected Chrome driver.
"""
import json

from requests import PreparedRequest, Response


class UCResponse(Response):
    """
    Response object for the undetected Chrome driver.

    Overrided to change some attributes and methods
    """

    def __init__(self):
        """
        NOTE:
            Response properties are:
            _content
            status_code
            headers
            url
            history
            encoding
            reason
            cookies
            elapsed
            request
        """
        super().__init__()
        self._content = None
        self.raw = ''
        self.reason = ''
        self.url = ''
        self.request = PreparedRequest()
        self.connection = None
        self.encoding = 'utf-8'
        self.status_code = 403

    def json(self, **kwargs):
        """
        Return the json content of the response.
        """
        try:
            # the following could raise a json.JSONDecodeError if the content is already a string
            json_content = super().json(**kwargs)
        except json.JSONDecodeError:
            json_content = json.loads(self._content)
        return json_content

    @property
    def content(self):
        """ Return the content of the response. """
        return self._content

    @content.setter
    def content(self, value):
        """ Setter for content. """
        self._content = value



# coding=utf-8
"""
Implementation for:
- AppCore: handle most of the lower level interaction with the downloader, lfs, and api components to make writing CLI/GUI code easier and cleaner and avoid duplication.
"""
import datetime
import json
import logging
import os
import shutil
from base64 import b64decode
from hashlib import sha1
from locale import getlocale, LC_CTYPE
from multiprocessing import Queue
from platform import system
from typing import List, Optional
from urllib.parse import urlparse

from requests import session
from requests.exceptions import ConnectionError, HTTPError

import UEVaultManager.tkgui.modules.globals as gui_g  # using the shortest variable name for globals for convenience
# noinspection PyPep8Naming
from UEVaultManager import __version__ as UEVM_version
from UEVaultManager.api.egs import EPCAPI
from UEVaultManager.api.uevm import UEVMAPI
from UEVaultManager.downloader.mp.DLManagerClass import DLManager
from UEVaultManager.lfs.EPCLFSClass import EPCLFS
from UEVaultManager.lfs.UEVMLFSClass import UEVMLFS
from UEVaultManager.lfs.utils import clean_filename, path_join
from UEVaultManager.models.Asset import Asset, InstalledAsset
from UEVaultManager.models.downloading import AnalysisResult, ConditionCheckResult
from UEVaultManager.models.exceptions import InvalidCredentialsError
from UEVaultManager.models.json_manifest import JSONManifest
from UEVaultManager.models.manifest import Manifest
from UEVaultManager.models.types import DateFormat
from UEVaultManager.tkgui.modules.functions import exit_and_clean_windows
from UEVaultManager.tkgui.modules.functions_no_deps import format_size
from UEVaultManager.utils.cli import check_and_create_file, check_and_create_folder
from UEVaultManager.utils.egl_crypt import decrypt_epic_data
from UEVaultManager.utils.env import is_windows_mac_or_pyi


# make some properties of the AppCore class accessible from outside to limit the number of imports needed
def log_info_and_gui_display(message: str) -> None:
    """
    Wrapper to log a message using a log function AND use a DisplayWindows to display the message if the gui is active.
    :param message: message to log.

    Notes:
       To avoid "PicklingError" when using multithreading, this method must still here, at the start of the file and outside the class (WTF ?)
       Do not use self. logger instead of logging.getLogger('Core')
    """
    # self.logger.info(message)
    logging.getLogger('Core').info(message)
    if gui_g.WindowsRef.display_content is not None:
        gui_g.WindowsRef.display_content.display(message)


class AppCore:
    """
    AppCore handles most of the lower level interaction with
    the downloader, lfs, and api components to make writing CLI/GUI
    code easier and cleaner and avoid duplication.
    :param override_config: path to a config file to use instead of the default.
    :param timeout: timeout for the request. Could be a float or a tuple of float (connect timeout, read timeout).
    """
    _egl_version = '11.0.1-14907503+++Portal+Release-Live'

    def __init__(self, override_config=None, timeout=(7, 7)):
        self.timeout = timeout
        self.logger = logging.getLogger('Core')
        self.egs = EPCAPI(timeout=self.timeout)
        self.uevmlfs = UEVMLFS(config_file=override_config)
        self.egl = EPCLFS()
        self.uevm_api = UEVMAPI()

        # on non-Windows load the programdata path from config
        if os.name != 'nt':
            self.egl.programdata_path = self.uevmlfs.config.get('UEVaultManager', 'egl_programdata', fallback=None)
            if self.egl.programdata_path and not os.path.exists(self.egl.programdata_path):
                self.logger.error(f'Config EGL path ("{self.egl.programdata_path}") is invalid! Disabling sync...')
                self.egl.programdata_path = None
                self.uevmlfs.config.remove_option('UEVaultManager', 'egl_programdata')
                self.uevmlfs.save_config()

        self.local_timezone = datetime.datetime.now().astimezone().tzinfo
        self.language_code, self.country_code = ('en', 'US')

        if locale := self.uevmlfs.config.get('UEVaultManager', 'locale', fallback=getlocale(LC_CTYPE)[0]):
            try:
                self.language_code, self.country_code = locale.split('-' if '-' in locale else '_')
                self.logger.debug(f'Set locale to {self.language_code}-{self.country_code}')
                # adjust egs api language as well
                self.egs.language_code, self.egs.country_code = self.language_code, self.country_code
            except Exception as error:
                self.logger.warning(f'Getting locale failed: {error!r}, falling back to using en-US.')
        elif system() != 'Darwin':  # macOS doesn't have a default locale we can query
            self.logger.warning('Could not determine locale, falling back to en-US')

        self.update_available = False
        self.force_show_update = False
        self.webview_killswitch = False
        self.user_is_connected = False

        # UE assets metadata cache properties
        self.ue_assets_count = 0
        # set to True to add print more information during long operations
        self.verbose_mode = False
        # Create a backup of the output file (when using the --output option) suffixed by a timestamp before creating a new file
        self.create_output_backup = True
        # Set the file name (and path) to log issues when an asset is ignored or filtered when running the list or scrap commands
        self.ignored_assets_filename_log = ''
        # Set the file name (and path) to log issues when an asset is not found on the marketplace when running the list or scrap commands
        self.notfound_assets_filename_log = ''
        # Set the file name (and path) to log issues when scanning folder to find assets
        self.scan_assets_filename_log = ''
        # Set the file name (and path) to log scraped assets
        self.scrap_assets_filename_log = ''
        # Create a backup of the log files that store asset analysis suffixed by a timestamp before creating a new file
        self.create_log_backup = True
        # file loggers
        self.ignored_logger = None
        self.notfound_logger = None
        self.scan_assets_logger = None
        self.scrap_asset_logger = None
        self.use_threads = False
        self.thread_executor = None
        self.thread_executor_must_stop = False
        self.engine_version_for_obsolete_assets = gui_g.s.engine_version_for_obsolete_assets
        # self.display_windows= DisplayContentWindow(title='UEVM command output', quit_on_close=True)
        self.display_windows = None
        self.session_ttl: int = 100  # time to live for the current user login session in seconds

    @staticmethod
    def load_manifest(data: bytes) -> Manifest:
        """
        Load a manifest.
        :param data: bytes object to load the manifest from.
        :return: Manifest object.
        """
        if data[0:1] == b'{':
            return JSONManifest.read_all(data)
        else:
            return Manifest.read_all(data)

    @staticmethod
    def check_installation_conditions(analysis: AnalysisResult, folders: [], ignore_space_req: bool = False) -> ConditionCheckResult:
        """
        Check installation conditions.
        :param analysis: analysis result to check.
        :param folders: folders to check free size for.
        :param ignore_space_req: whether to ignore space requirements or not.
        :return: ConditionCheckResult object.
        """
        results = ConditionCheckResult(failures=set(), warnings=set())
        if not isinstance(folders, list):
            folders = [folders]
        for folder in folders:
            if not folder:
                results.failures.add(f'"At least one folder is not defined. Check your config and command options.')
                break
            if not os.path.exists(folder):
                results.failures.add(
                    f'"{folder}" does not exist. Check your config and command options and make sure all necessary disks are available.'
                )
                break
            min_disk_space = analysis.disk_space_delta
            _, _, free = shutil.disk_usage(folder)
            if free < min_disk_space:
                free_gib = free / 1024 ** 3
                required_gib = min_disk_space / 1024 ** 3
                message = f'"{folder}": Potentially not enough available disk space: {free_gib:.02f} GiB < {required_gib:.02f} GiB'
                if ignore_space_req:
                    results.warnings.add(message)
                else:
                    results.failures.add(message)
        return results

    def setup_assets_loggers(self) -> None:
        """
        Setup logging for ignored, not found and bad data assets.
        """

        def create_logger(logger_name: str, filename_log: str):
            """
            Create a logger for ignored, not found and bad data assets.
            :param logger_name: logger name.
            :param filename_log: log file name.
            :return: logger.
            """
            filename_log = filename_log.replace('~/.config', self.uevmlfs.path)
            if check_and_create_file(filename_log):
                handler = logging.FileHandler(filename_log, mode='w')
                handler.setFormatter(formatter)
                logger = logging.Logger(logger_name, 'INFO')
                logger.addHandler(handler)
                logger.info(message)
                return logger
            else:
                self.logger.warning(f'Failed to create logger for file: {filename_log}')
                return None

        formatter = logging.Formatter('%(message)s')
        message = f"-----\n{datetime.datetime.now().strftime(DateFormat.csv)} Log Started\n-----\n"

        if self.ignored_assets_filename_log:
            self.ignored_logger = create_logger('IgnoredAssets', self.ignored_assets_filename_log)
        if self.notfound_assets_filename_log:
            self.notfound_logger = create_logger('NotFoundAssets', self.notfound_assets_filename_log)
        if self.scan_assets_filename_log:
            self.scan_assets_logger = create_logger('ScanAssets', self.scan_assets_filename_log)
        if self.scrap_assets_filename_log:
            self.scrap_asset_logger = create_logger('ScrapAssets', self.scrap_assets_filename_log)

    def auth_sid(self, sid) -> str:
        """
        Handles getting an exchange code from an id.
        :param sid: session id.
        :return: exchange code.
        """
        s = session()
        s.headers.update(
            {
                'X-Epic-Event-Action':
                'login',
                'X-Epic-Event-Category':
                'login',
                'X-Epic-Strategy-Flags':
                '',
                'X-Requested-With':
                'XMLHttpRequest',
                'User-Agent':
                'Mozilla/5.0 (Windows NT 10.0; Win64; x64) '
                'AppleWebKit/537.36 (KHTML, like Gecko) '
                f'EpicGamesLauncher/{self._egl_version} '
                'UnrealEngine/4.23.0-14907503+++Portal+Release-Live '
                'Chrome/84.0.4147.38 Safari/537.36'
            }
        )
        s.cookies['EPIC_COUNTRY'] = self.country_code.upper()

        # get first set of cookies (EPIC_BEARER_TOKEN etc.)
        _ = s.get('https://www.epicgames.com/id/api/set-sid', params=dict(sid=sid))
        # get XSRF-TOKEN and EPIC_SESSION_AP cookie
        _ = s.get('https://www.epicgames.com/id/api/csrf')
        # finally, get the exchange code
        r = s.post('https://www.epicgames.com/id/api/exchange/generate', headers={'X-XSRF-TOKEN': s.cookies['XSRF-TOKEN']})

        if r.status_code == 200:
            return r.json()['code']

        self.logger.error(f'Getting exchange code failed: {r.json()}')
        return ''

    def auth_code(self, code) -> bool:
        """
        Handles authentication via authorization code (either retrieved manually or automatically).
        """
        try:
            self.uevmlfs.userdata = self.egs.start_session(authorization_code=code)
            return True
        except Exception as error:
            self.logger.error(f'Log in failed with {error!r}, please try again.')
            return False

    def auth_ex_token(self, code) -> bool:
        """
        Handles authentication via exchange token (either retrieved manually or automatically).
        """
        try:
            self.uevmlfs.userdata = self.egs.start_session(exchange_token=code)
            return True
        except Exception as error:
            self.logger.error(f'Log in failed with {error!r}, please try again.')
            return False

    def auth_import(self) -> bool:
        """
        Import refresh token from EGL installation and use it to log in.
        :return: True if successful, False otherwise.
        """
        remember_me_data = self.egl.config.get('RememberMe', 'Data')
        raw_data = b64decode(remember_me_data)
        # data is encrypted
        if raw_data[0] != '{':
            for data_key in self.egl.data_keys:
                try:
                    decrypted_data = decrypt_epic_data(data_key, raw_data)
                    re_data = json.loads(decrypted_data)[0]
                    break
                except Exception as error:
                    self.logger.debug(f'Decryption with key {data_key} failed with {error!r}')
            else:
                raise ValueError('Decryption of EPIC launcher user information failed.')
        else:
            re_data = json.loads(raw_data)[0]

        if 'Token' not in re_data:
            raise ValueError('No login session in config')
        refresh_token = re_data['Token']
        try:
            self.uevmlfs.userdata = self.egs.start_session(refresh_token=refresh_token)
            return True
        except Exception as error:
            self.logger.error(f'Logging failed with {error!r}, please try again.')
            return False

    def login(self, force_refresh: bool = False, raise_error: bool = False) -> bool:
        """
        Attempt log in with existing credentials.
        :param force_refresh: whether to force a refresh of the session.
        :param raise_error: whether to raise an exception if login fails.
        :return: True if successful, False otherwise.
        """
        if not self.uevmlfs.userdata:
            if raise_error:
                raise ValueError('No saved credentials')
            else:
                self.user_is_connected = False
                return False
        elif self.user_is_connected and self.uevmlfs.userdata['expires_at']:
            dt_exp = datetime.datetime.fromisoformat(self.uevmlfs.userdata['expires_at'][:-1])
            dt_now = datetime.datetime.now(datetime.UTC)
            td = dt_now.timestamp() - dt_exp.timestamp()
            if td > self.session_ttl:
                return True
            else:
                self.user_is_connected = False

        # run update check
        if self.update_check_enabled():
            try:
                self.check_for_updates()
            except Exception as error:
                self.logger.warning(f'Checking for UEVaultManager updates failed: {error!r}')

        if self.uevmlfs.userdata['expires_at'] and not force_refresh:
            dt_exp = datetime.datetime.fromisoformat(self.uevmlfs.userdata['expires_at'][:-1])
            dt_now = datetime.datetime.now(datetime.UTC)
            td = dt_now.timestamp() - dt_exp.timestamp()

            if td > self.session_ttl:
                self.logger.info('Trying to re-use existing login session...')
                try:
                    self.egs.resume_session(self.uevmlfs.userdata)
                    self.user_is_connected = True
                    return True
                except InvalidCredentialsError as error:
                    self.logger.warning(f'Resuming failed due to invalid credentials: {error!r}')
                except (Exception, ) as error:
                    self.logger.warning(f'Resuming failed for unknown reason: {error!r}')
                # If verify fails just continue the normal authentication process
                self.logger.info('Falling back to using refresh token...')

        try:
            self.logger.info('Logging in...')
            userdata = self.egs.start_session(self.uevmlfs.userdata['refresh_token'])
        except InvalidCredentialsError as error:
            self.logger.warning(f'Resuming failed due to invalid credentials: {error!r}')
            self.uevmlfs.invalidate_userdata()
            return False
        except (Exception, ) as error:
            self.logger.error(f'Connection failed with error : {error!r}')
            return False
        except (HTTPError, ConnectionError) as error:
            self.logger.error(f'HTTP request for log in failed: {error!r}, please try again later.')
            return False

        self.uevmlfs.userdata = userdata
        self.user_is_connected = True
        return True

    def update_check_enabled(self) -> bool:
        """
        Return whether update checks are enabled or not.
        :return: True if update checks are enabled, False otherwise.
        """
        return not self.uevmlfs.config.getboolean('UEVaultManager', 'disable_update_check', fallback=False)

    def update_notice_enabled(self) -> bool:
        """
        Return whether update notices are enabled or not.
        :return: True if update notices are enabled, False otherwise.
        """
        if self.force_show_update:
            return True
        return not self.uevmlfs.config.getboolean('UEVaultManager', 'disable_update_notice', fallback=not is_windows_mac_or_pyi())

    def check_for_updates(self, force=False) -> None:
        """
        Check for updates and sets the update_available flag accordingly.
        :param force: force update check.
        """

        def version_tuple(v):
            """
            Convert a version string to a tuple of ints.
            :param v: version string.
            :return:  tuple of int.
            """
            return tuple(map(int, (v.split('.'))))

        cached = self.uevmlfs.get_online_version_saved()
        version_info = cached['data']
        if force or not version_info or (datetime.datetime.now().timestamp() - cached['last_update']) > 24 * 3600:
            version_info = self.uevm_api.get_online_version_information()
            self.uevmlfs.set_online_version_saved(version_info)

        web_version = version_info['version']
        self.update_available = version_tuple(web_version) > version_tuple(UEVM_version)

    def get_egl_version(self):
        """
        return the egl version.
        :return:
        """
        return self._egl_version

    def get_update_info(self) -> dict:
        """
        Return update info dict.
        :return: update info dict.
        """
        return self.uevmlfs.get_online_version_saved()['data']

    def is_installed(self, app_name: str) -> bool:
        """
        Return whether an asset is installed.
        :param app_name: asset name to check.
        :return: True if asset is installed, False otherwise.
        """
        return self.uevmlfs.get_installed_asset(app_name) is not None

    def get_installed_manifest(self, app_name):
        """
        Get the installed manifest.
        :param app_name: asset name to get the installed manifest for.
        :return:
        """
        installed_asset = self.uevmlfs.get_installed_asset(app_name)
        old_bytes = self.uevmlfs.load_manifest(app_name, installed_asset.version, installed_asset.platform)
        return old_bytes, installed_asset.base_urls

    def asset_obj_from_json(self, app_name: str) -> Optional[Asset]:
        """
        return an "item" like for compatibilty with "old" methods .
        :param app_name: asset name to get the asset object for.
        :return: Asset object.
        """
        asset_data, _ = self.uevmlfs.get_asset(app_name)
        if asset_data:
            return Asset.from_json(asset_data)  # create an object from the Asset class using the json data
        return None

    def get_cdn_urls(self, item, platform: str = 'Windows'):
        """
        Get the CDN URLs.
        :param item: item to get the CDN URLs for.
        :param platform: platform to get the CDN URLs for.
        :return: list of CDN URLs.
        """
        m_api_r = self.egs.get_item_manifest(item.namespace, item.catalog_item_id, item.app_name, platform)

        # never seen this outside the launcher itself, but if it happens: PANIC!
        if len(m_api_r['elements']) > 1:
            raise ValueError('Manifest response has more than one element!')

        manifest_hash = m_api_r['elements'][0]['hash']
        base_urls = []
        manifest_urls = []
        for manifest in m_api_r['elements'][0]['manifests']:
            base_url = manifest['uri'].rpartition('/')[0]
            if base_url not in base_urls:
                base_urls.append(base_url)

            if 'queryParams' in manifest:
                params = '&'.join(f'{p["name"]}={p["value"]}' for p in manifest['queryParams'])
                manifest_urls.append(f'{manifest["uri"]}?{params}')
            else:
                manifest_urls.append(manifest['uri'])

        return manifest_urls, base_urls, manifest_hash

    def get_cdn_manifest(self, item, platform: str = 'Windows', disable_https=False):
        """
        Get the CDN manifest.
        :param item: item to get the CDN manifest for.
        :param platform: platform to get the CDN manifest for.
        :param disable_https: disable HTTPS for the manifest URLs.
        :return: tuple (manifest data, base URLs, request status code).
        """
        manifest_urls, base_urls, manifest_hash = self.get_cdn_urls(item, platform)
        if not manifest_urls:
            raise ValueError('No manifest URLs returned by API')

        if disable_https:
            manifest_urls = [url.replace('https://', 'http://') for url in manifest_urls]

        r = {}
        for url in manifest_urls:
            self.logger.debug(f'Trying to download manifest from "{url}"...')
            try:
                r = self.egs.unauth_session.get(url, timeout=self.timeout)
            except Exception as error:
                self.logger.warning(f'Failed to download manifest from "{urlparse(url).netloc}" (Exception: {error!r}), trying next URL...')
                continue

            if r.status_code == 200:
                manifest_bytes = r.content
                break
            else:
                self.logger.warning(f'Failed to download manifest from "{urlparse(url).netloc}" (status: {r.status_code}), trying next URL...')
        else:
            raise ValueError(f'Failed to get manifest from any CDN URL, last result: {r.status_code} ({r.reason})')

        if sha1(manifest_bytes).hexdigest() != manifest_hash:
            raise ValueError('Manifest sha hash mismatch!')

        return manifest_bytes, base_urls, r.status_code

    def get_uri_manifest(self, uri: str) -> (bytes, List[str]):
        """
        Get the manifest.
        :param uri: uRI to get the manifest from.
        :return:  Manifest data and base URLs.
        """
        if uri.startswith('http'):
            r = self.egs.unauth_session.get(uri)
            r.raise_for_status()
            new_manifest_data = r.content
            base_urls = [r.url.rpartition('/')[0]]
        else:
            base_urls = []
            with open(uri, 'rb') as f:
                new_manifest_data = f.read()

        return new_manifest_data, base_urls

    def prepare_download(
        self,
        base_asset: Asset,  # contains generic info of the base asset for all releases, NOT the selected release
        release_name: str,
        release_title: str,
        download_folder: str = '',
        install_folder: str = '',
        no_resume: bool = False,
        platform: str = 'Windows',
        max_shm: int = 0,
        max_workers: int = 0,
        dl_optimizations: bool = False,
        override_manifest: str = '',
        override_old_manifest: str = '',
        override_base_url: str = '',
        status_queue: Queue = None,
        reuse_last_install: bool = False,
        disable_patching: bool = False,
        file_prefix_filter: list = None,
        file_exclude_filter: list = None,
        file_install_tag: list = None,
        preferred_cdn: str = None,
        disable_https: bool = False
    ) -> (DLManager, AnalysisResult, InstalledAsset):
        """
        Prepare a download.
        :param base_asset: "base" asset to prepare the download for, not the selected release.
        :param release_name: release name prepare the download for.
        :param release_title: release title prepare the download for.
        :param download_folder: folder to download the asset to.
        :param install_folder: base folder to install the asset to.
        :param platform: platform to prepare the download for.
        :param no_resume: avoid to resume. Force a new download.
        :param max_shm: maximum amount of shared memory to use.
        :param max_workers: maximum number of workers to use.
        :param dl_optimizations: download optimizations.
        :param override_manifest: override the manifest.
        :param override_old_manifest: override the old manifest.
        :param override_base_url: override the base URL.
        :param reuse_last_install: update previous installation.
        :param disable_patching: disable patching.
        :param status_queue: status queue to send status updates to.
        :param file_prefix_filter: file prefix filter.
        :param file_exclude_filter: file exclude filter.
        :param file_install_tag: file install tag.
        :param preferred_cdn: preferred CDN.
        :param disable_https: disable HTTPS. For LAN installs only.
        :return: (DLManager object, AnalysisResult object, InstalledAsset object).
        """
        old_manifest = None
        egl_guid = ''

        # load old manifest if we have one
        if override_old_manifest:
            log_info_and_gui_display(f'Overriding old manifest with "{override_old_manifest}"')
            old_bytes, _ = self.get_uri_manifest(override_old_manifest)
            old_manifest = self.load_manifest(old_bytes)
        elif not disable_patching and not no_resume and self.is_installed(release_name):
            old_bytes, _base_urls = self.get_installed_manifest(release_name)
            if _base_urls and not base_asset.base_urls:
                base_asset.base_urls = _base_urls
            if not old_bytes:
                self.logger.error(f'Could not load old manifest, patching will not work!')
            else:
                old_manifest = self.load_manifest(old_bytes)

        base_urls = base_asset.base_urls

        # The EGS client uses plaintext HTTP by default for the purposes of enabling simple DNS based
        # CDN redirection to a (local) cache.
        disable_https = disable_https or self.uevmlfs.config.getboolean('UEVaultManager', 'disable_https', fallback=False)

        if override_manifest:
            log_info_and_gui_display(f'Overriding manifest with "{override_manifest}"')
            new_manifest_data, _base_urls = self.get_uri_manifest(override_manifest)
            # if override manifest has a base URL use that instead
            if _base_urls:
                base_urls = _base_urls
        else:
            new_manifest_data, base_urls, status_code = self.get_cdn_manifest(base_asset, platform, disable_https=disable_https)
            # overwrite base urls in metadata with current ones to avoid using old/dead CDNs
            base_asset.base_urls = base_urls

        if not new_manifest_data:
            message = f'Manifest data is empty for "{release_name}". Could be a timeout or an issue with the connection.'
            raise ValueError(message)

        log_info_and_gui_display('Parsing game manifest...')
        manifest = self.load_manifest(new_manifest_data)
        self.logger.debug(f'Base urls: {base_urls}')
        # save manifest with version name as well for testing/downgrading/etc.
        manifest_filename = self.uevmlfs.save_manifest(release_name, new_manifest_data, version=manifest.meta.build_version, platform=platform)

        # make sure donwnload folder actually exists (but do not create asset folder)
        if not check_and_create_folder(download_folder):
            log_info_and_gui_display(f'"{download_folder}" did not exist, it has been created.')
        if not os.access(download_folder, os.W_OK):
            raise PermissionError(f'No write access to "{download_folder}"')

        # reuse existing installation's directory
        installed_asset = self.uevmlfs.get_installed_asset(release_name)
        if reuse_last_install and installed_asset:
            install_path = installed_asset.install_path
            egl_guid = installed_asset.egl_guid
        else:
            # asset are always installed in the 'Content' sub folder
            # NO we don't want to store "content" in the "install path"
            # install_path = path_join(install_folder, 'Content') if install_folder  else ''
            install_path = install_folder

        # check for write access on the installation path or its parent directory if it doesn't exist yet
        if install_path:
            log_info_and_gui_display(f'Install path: {install_path}')
            if not check_and_create_folder(install_path):
                log_info_and_gui_display(f'"{install_path}" did not exist, it has been created.')
            if not os.access(install_path, os.W_OK):
                raise PermissionError(f'No write access to "{install_path}"')

        if not no_resume:
            filename = clean_filename(f'{release_name}.resume')
            resume_file = path_join(self.uevmlfs.tmp_folder, filename)
        else:
            resume_file = None

        # Use user-specified base URL or preferred CDN first, otherwise fall back to
        # EGS's behaviour of just selecting the first CDN in the list.
        base_url = None
        if override_base_url:
            log_info_and_gui_display(f'Overriding base URL with "{override_base_url}"')
            base_url = override_base_url
        elif preferred_cdn or (preferred_cdn := self.uevmlfs.config.get('UEVaultManager', 'preferred_cdn', fallback=None)):
            for url in base_urls:
                if preferred_cdn in url:
                    base_url = url
                    break
            else:
                self.logger.warning(f'Preferred CDN "{preferred_cdn}" unavailable, using default selection.')
        # Use first, fail if none known
        if not base_url:
            if not base_urls:
                raise ValueError('No base URLs found, please try again.')
            base_url = base_urls[0]

        if disable_https:
            base_url = base_url.replace('https://', 'http://')

        self.logger.debug(f'Using base URL: {base_url}')
        scheme, cdn_host = base_url.split('/')[0:3:2]
        log_info_and_gui_display(f'Selected CDN: {cdn_host} ({scheme.strip(":")})')

        if not max_shm:
            max_shm = self.uevmlfs.config.getint('UEVaultManager', 'max_memory', fallback=2048)

        if dl_optimizations:
            log_info_and_gui_display('Download order optimizations are enabled.')
            process_opt = True
        else:
            process_opt = False

        if not max_workers:
            max_workers = self.uevmlfs.config.getint('UEVaultManager', 'max_workers', fallback=0)

        download_manager = DLManager(
            download_dir=download_folder,
            base_url=base_url,
            resume_file=resume_file,
            status_q=status_queue,
            max_shared_memory=max_shm * 1024 * 1024,
            max_workers=max_workers,
            timeout=self.timeout,
            trace_func=log_info_and_gui_display,
        )
        installed_asset = self.uevmlfs.get_installed_asset(release_name)
        if installed_asset is None:
            # create a new installed asset
            installed_asset = InstalledAsset(app_name=release_name, title=release_title)
        # update the installed asset
        installed_asset.version = manifest.meta.build_version
        installed_asset.base_urls = base_urls
        installed_asset.egl_guid = egl_guid
        installed_asset.manifest_path = override_manifest if override_manifest else manifest_filename
        installed_asset.platform = platform
        installed_asset.catalog_item_id = base_asset.catalog_item_id
        already_installed = install_path and install_path in installed_asset.installed_folders
        analyse_res = download_manager.run_analysis(
            manifest=manifest,
            old_manifest=old_manifest,
            patch=not disable_patching,
            resume=not no_resume,
            file_prefix_filter=file_prefix_filter,
            file_exclude_filter=file_exclude_filter,
            file_install_tag=file_install_tag,
            processing_optimization=process_opt,
            already_installed=already_installed
        )
        if install_path:
            # will add install_path to the installed_folders list after checking if it is not already in it
            installed_asset.install_path = install_path
        installed_asset.install_size = analyse_res.install_size
        return download_manager, analyse_res, installed_asset

    def clean_exit(self, code=0) -> None:
        """
        Do cleanup, config saving, and quit.
        :param code: exit code.
        """
        self.uevmlfs.save_config()
        logging.shutdown()
        exit_and_clean_windows(code)

    def open_manifest_file(self, file_path: str) -> dict:
        """
        Open a manifest file and return its data.
        :param file_path: path to the manifest file.
        :return: manifest data.
        """
        try:
            with open(file_path, 'rb') as file:
                manifest_data = file.read()
        except FileNotFoundError:
            self.logger.warning(f'The file {file_path} does not exist.')
            return {}
        manifest_info = {}
        manifest = self.load_manifest(manifest_data)
        manifest_info['app_name'] = manifest.meta.app_name

        # file and chunk count
        manifest_info['num_files'] = manifest.file_manifest_list.count
        manifest_info['num_chunks'] = manifest.chunk_data_list.count
        # total file size
        total_size = sum(fm.file_size for fm in manifest.file_manifest_list.elements)
        file_size = format_size(total_size)
        manifest_info['file_size'] = file_size
        manifest_info['disk_size'] = total_size
        # total chunk size
        total_size = sum(c.file_size for c in manifest.chunk_data_list.elements)
        chunk_size = format_size(total_size)
        manifest_info['chunk_size'] = chunk_size
        manifest_info['download_size'] = total_size
        return manifest_info



# coding: utf-8
"""
Implementation for:
- EPCAPI : Epic Games Client API
- create_empty_assets_extra : Create an empty asset extra dict.
- is_asset_obsolete : Check if an asset is obsolete.
"""
import json
import logging
from pathlib import Path
from typing import cast

import nodriver as uc
import requests
import requests.adapters
from bs4 import BeautifulSoup
from requests import Response
from requests.adapters import HTTPAdapter
from requests.auth import HTTPBasicAuth
from requests.utils import dict_from_cookiejar
from selenium.webdriver.remote.command import Command

from UEVaultManager.models.exceptions import InvalidCredentialsError
from UEVaultManager.models.UCRequest import UCRequest
from UEVaultManager.models.UCResponse import UCResponse
from UEVaultManager.tkgui.modules.types import GrabResult, UCRequestType
from UEVaultManager.utils.cli import create_list_from_string


def is_asset_obsolete(supported_versions='', engine_version_for_obsolete_assets=None) -> bool:
    """
    :param supported_versions: supported versions the check the obsolete status against.
    :param engine_version_for_obsolete_assets: engine version to use to check if an asset is obsolete.
    :return: True if the asset is obsolete, False otherwise.
    """
    if not engine_version_for_obsolete_assets or not supported_versions:
        obsolete = False
    else:
        supported_versions_list = supported_versions.lower().replace('ue_', '').replace('ea', '')
        supported_versions_list = create_list_from_string(supported_versions_list)
        obsolete = True
        for _, version in enumerate(supported_versions_list):
            if not version:
                continue
            else:
                if float(engine_version_for_obsolete_assets) <= float(version):
                    obsolete = False
                    break
    return obsolete


def create_empty_assets_extra(asset_name: str) -> dict:
    """
    Create an empty asset extra dict.
    :param asset_name:  The name of the asset.
    :return: empty asset extra dict.
     """
    return {
        'asset_name': asset_name,
        'asset_slug': '',
        'price': 0,
        'discount_price': 0,
        'review': 0,
        'owned': True,
        'discount_percentage': 0,
        'discounted': False,
        'asset_url': '',
        'page_title': '',
        # 'supported_versions': '',
        'installed_folders': [],
        'grab_result': GrabResult.NO_ERROR.name,
    }


class EPCAPI:
    """
    Epic Games Client API.
    :param lc: language code.
    :param cc: country code.
    :param timeout: timeout for the request. Could be a float or a tuple of float (connect timeout, read timeout).
    """
    notfound_logger = None
    scrap_asset_logger = None

    _user_agent = 'UELauncher/11.0.1-14907503+++Portal+Release-Live Windows/10.0.19041.1.256.64bit'
    # _user_agent = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36'
    # _store_user_agent = 'EpicGamesLauncher/14.0.8-22004686+++Portal+Release-Live'
    # required for the oauth request
    # _label = 'Live-EternalKnight'
    _user_basic = '34a02cf8f4414e29b15921876da36f9a'
    _pw_basic = 'daafbccc737745039dffe53d94fc76cf'

    _oauth_host = 'https://account-public-service-prod03.ol.epicgames.com'
    _launcher_host = 'https://launcher-public-service-prod06.ol.epicgames.com'
    # _entitlements_host = 'entitlement-public-service-prod08.ol.epicgames.com'
    _catalog_host = 'https://catalog-public-service-prod06.ol.epicgames.com'
    # _ecommerce_host = 'https://ecommerceintegration-public-service-ecomprod02.ol.epicgames.com'
    # _datastorage_host = 'https://datastorage-public-service-liveegs.live.use1a.on.epicgames.com'
    # _library_host = 'https://library-service.live.use1a.on.epicgames.com'
    # Using the actual store host with a user-agent newer than 14.0.8 leads to a CF verification page,
    # but the dedicated graphql host works fine.
    # _store_gql_host = 'https://launcher.store.epicgames.com'
    # _store_gql_host = 'https://graphql.epicgames.com'
    # _artifact_service_host = 'https://artifact-public-service-prod.beee.live.use1a.on.epicgames.com'
    # _login_url = 'https://www.unrealengine.com/id/login/epic'

    # THESE URLs COULD NOT BE CALL USING SESSION ANYMORE DU TO RECATCHA VALIDATION
    _url_marketplace = 'https://www.unrealengine.com/marketplace'
    _url_asset_list = f'{_url_marketplace}/api/assets'
    _url_owned_assets = f'{_url_asset_list}/vault'
    _url_asset = f'{_url_asset_list}/asset'
    _url_search_asset = f'{_url_marketplace}/en-US/product'

    # page d'un asset avec son urlSlug
    # _url_marketplace/en-US/product/{'urlSlug}
    # https://www.unrealengine.com/marketplace/en-US/product/volcrate
    #
    # detail json d'un asset avec son id (et non pas son asset_id ou son catalog_id)
    # UE_ASSET/{el['id']}")
    # https://www.unrealengine.com/marketplace/api/assets/asset/d27cf128fdc24e328cf950b019563bc5
    #
    # liste json des reviews d'un asset avec son id
    # https://www.unrealengine.com/marketplace/api/review/d27cf128fdc24e328cf950b019563bc5/reviews/list?start=0&count=10&sortBy=CREATEDAT&sortDir=DESC
    #
    # liste json des questions d'un asset avec son id
    # https://www.unrealengine.com/marketplace/api/review/d27cf128fdc24e328cf950b019563bc5/questions/list?start=0&count=10&sortBy=CREATEDAT&sortDir=DESC
    #
    # liste json des tags courants
    # https://www.unrealengine.com/marketplace/api/tags
    """ 
    le champ release_info contient l'ID des manifest à telecharger pour chaque version
    (voir app_id) 
    urls dans EAM
    - start_session
    https://account-public-service-prod03.ol.epicgames.com/account/api/oauth/token
    
    - resume_session
    https://account-public-service-prod03.ol.epicgames.com/account/api/oauth/verify
    
    - invalidate_sesion
    https://account-public-service-prod03.ol.epicgames.com/account/api/oauth/sessions/kill/{}", access_token);
    
    - account_details
    https://account-public-service-prod03.ol.epicgames.com/account/api/public/account/{}
    
    - account_ids_details
    https://account-public-service-prod03.ol.epicgames.com/account/api/public/account
    
    - account_friends
    https://friends-public-service-prod06.ol.epicgames.com/friends/api/public/friends/{}?includePending={}", id, include_pending);
    
    # asset
    https://launcher-public-service-prod06.ol.epicgames.com/launcher/api/public/assets/{}?label={}", plat, lab);
    
    - asset_manifest
    https://launcher-public-service-prod06.ol.epicgames.com/launcher/api/public/assets/v2/platform/{}/namespace/{}/catalogItem/{}/app/{}/label/{}",
    
    - asset_info
    https://catalog-public-service-prod06.ol.epicgames.com/catalog/api/shared/namespace/{}/bulk/items?id={}&includeDLCDetails=true&includeMainGameDetails=true&country=us&locale=lc",asset.namespace, asset.catalog_item_id);
    
    - game_token
    https://account-public-service-prod03.ol.epicgames.com/account/api/oauth/exchange"
    
    - ownership_token
    https://ecommerceintegration-public-service-ecomprod02.ol.epicgames.com/ecommerceintegration/api/public/platforms/EPIC/identities/{}/ownershipToken",
    
    - user_entitlements
    https://entitlement-public-service-prod08.ol.epicgames.com/entitlement/api/account/{}/entitlements?start=0&count=5000",
    
    - library_items
    https://library-service.live.use1a.on.epicgames.com/library/api/public/items?includeMetadata={}", include_metadata)
    https://library-service.live.use1a.on.epicgames.com/library/api/public/items?includeMetadata={}&cursor={}", include_metadata, c)
    """

    def __init__(self, lc='en', cc='US', timeout=(7, 7)):
        self.logger = logging.getLogger('EPCAPI')
        self.session = requests.session()
        self.session.headers['User-Agent'] = self._user_agent
        # increase maximum pool size for multithreaded metadata requests
        self.session.mount('https://', requests.adapters.HTTPAdapter(pool_maxsize=16))

        self.unauth_session = requests.session()
        self.unauth_session.headers['User-Agent'] = self._user_agent

        self._oauth_basic = HTTPBasicAuth(self._user_basic, self._pw_basic)

        self.access_token = None
        self.user = None

        self.language_code = lc
        self.country_code = cc

        self.timeout = timeout
        self._uc_browser = None
        self._uc_request = UCRequest()
        self.debug_mode = False
        # set to true to use "NoDriver" object instead of a usual session object to bypass captcha on marketplace pages
        self.bypass_captcha = False

    def extract_price(self, price_text=None, asset_name='Unknown') -> float:
        """
        Extracts the price from a string.
        :param price_text: string to extract the price from.
        :param asset_name: name of the asset (for display purpose only).
        :return: price.
        """
        if not price_text:
            self.logger.debug(f'Price not found for {asset_name}')
            return 0.0
        price = 0.0
        try:
            # asset base price when logged
            price = str(price_text)
            price = price.strip('$€')
            price = price.replace(',', '')
            price = float(price)
        except Exception as error:
            self.logger.warning(f'Can not find the price for {asset_name}:{error!r}')
        return price

    def get_scrap_url(self, start=0, count=1, sort_by='effectiveDate', sort_order='DESC') -> str:
        """
        Return the scraping URL for an asset.
        """
        url = f'{self._url_asset_list}?start={start}&count={count}&sortBy={sort_by}&sortDir={sort_order}'
        # other possible filters
        # to see the list of possible filters: https://www.unrealengine.com/marketplace/en-US/assets and use filters on the right panel.
        """
        # can add multiple platform filters
        # Windows, Android, Linux, Mac, PS4, Nintendo%20Switch, Win32, iOS, Xbox%20One, HTML5...
        url += f'&platform={platform}'

        # can add multiple compatibleWith filters
        url += f'&compatibleWith=UE_5.2'

        # can add multiple tag filters
        url += f'&tag=22'

        url += f'&discountPercentageRange={discount_percent}'
        url += f'&priceRange=%5B{prince_min*100}%2C{prince_max*100}'
        """
        return url

    def get_owned_scrap_url(self, start=0, count=1) -> str:
        """
        Return the scraping URL for an owned asset.
        """
        url = f'{self._url_owned_assets}?start={start}&count={count}'
        return url

    def get_marketplace_product_url(self, asset_slug: str = '') -> str:
        """
        Return the url for the asset in the marketplace.
        :param asset_slug: asset slug.
        :return: url.
        """
        url = f'{self._url_search_asset}/{asset_slug}'
        return url

    def get_api_product_url(self, uid: str = '') -> str:
        """
        Return the url for the asset using the UE API.
        :param uid: id of the asset (not the slug, nor the catalog_id).
        :return: url.
        """
        url = f'{self._url_asset}/{uid}'
        return url

    def get_owned_library(self) -> dict:
        """
        Get the owned library, INCLUDING the epic game assets owned
        :return: dict of assets.
        """
        platform = 'Windows'
        label = 'Live'
        json_data = {}
        try:
            r = self.session.get(f'{self._launcher_host}/launcher/api/public/assets/{platform}', params=dict(label=label), timeout=self.timeout)
            if r.ok:
                json_data = r.json()
        except Exception as error:
            self.logger.warning(f'Can not get the get owned library assets: {error!r}')
        return json_data

    def get_available_assets_count(self) -> int:
        """
        Return the number of assets in the marketplace.
        """
        url = self._url_asset_list
        r = self.get_url_with_uc(url, timeout=self.timeout, force_bypass_captcha=True)
        try:
            json_content = r.json()
            return int(json_content['data']['paging']['total'])
        except Exception as error:
            self.logger.warning(f'Can not get the asset count from {url}: {error!r}')
            return 0

    def is_valid_url(self, url='') -> bool:
        """
        Check is the url is valid (i.e. http response status is 200).
        :param url: url to check.
        :return: True if the url is valid.
        """
        if not url:
            return False

        try:
            r = self.get_url_with_uc(url, timeout=self.timeout)
        except (Exception, ):
            self.logger.warning(f'Timeout for {url}')
            raise ConnectionError()
        if r.status_code == 200:
            return True

    def get_json_data_from_url(self, url='', override_timeout=-1) -> dict:
        """
        Return the scraped assets.
        :param url: url to scrap.
        :param override_timeout: override the timeout set for the current object.
        :return: json data.

        Notes:
            Getting the data could take more time than other calls to the API. Use the override_timeout parameter to set a longer timeout if needed.
        """
        json_data = {}
        if not url:
            return json_data

        timeout = self.timeout if override_timeout == -1 else override_timeout
        r = self.get_url_with_uc(url, timeout=timeout)
        # print session content for debug
        # noinspection GrazieInspection
        # from pprint import pprint

        # try:
        #     session = self.session
        #     content = r.text
        #     print('\n\n===================\n\n')
        #     pprint(vars(session), depth=20, compact=True, width=120, sort_dicts=False)
        #     print('\n\n===================\n\n')
        #     print(url)
        #     print('\n\n===================\n\n')
        #     print(content)
        # except Exception as e:
        #     print(e)
        if r.status_code == 403:
            # probably a captcha page
            # make a second request to get the json data with captcha bypass
            self.logger.warning(f'403 Forbidden result for {url}. Testing a request with captcha bypass.')
            old_bypass_captcha = self.bypass_captcha
            self.bypass_captcha = True
            try:
                r = self.get_url_with_uc(url, timeout=timeout)
                if not r.ok:
                    self.logger.warning(f'Request with captcha bypass failed')
            except Exception as error:
                self.logger.warning(f'Request with captcha bypass failed. Error {error!r}')
                self.bypass_captcha = old_bypass_captcha

        return r.json() if r.ok else {}

    def resume_session(self, session: dict) -> dict:
        """
        Resumes a session.
        :param session: session to resume.
        :return: session.
        """
        self.session.headers['Authorization'] = f'bearer {session["access_token"]}'
        url = f'{self._oauth_host}/account/api/oauth/verify'
        r = self.session.get(url, timeout=self.timeout)
        if r.status_code >= 500:
            r.raise_for_status()

        j = r.json()
        if 'errorMessage' in j:
            self.logger.warning(f'Login to EGS API failed with errorCode: {j["errorCode"]}')
            raise InvalidCredentialsError(j['errorCode'])

        # update other data
        session.update(j)
        self.user = session
        return self.user

    def start_session(
        self, refresh_token: str = None, exchange_token: str = None, authorization_code: str = None, client_credentials: bool = False
    ) -> dict:
        """
        Start a session.
        :param refresh_token: refresh token.
        :param exchange_token: exchange token.
        :param authorization_code: authorization code.
        :param client_credentials: client credentials.
        :return: session.
        :raise: ValueError,InvalidCredentialsError.
        """
        if refresh_token:
            params = dict(grant_type='refresh_token', refresh_token=refresh_token, token_type='eg1')
        elif exchange_token:
            params = dict(grant_type='exchange_code', exchange_code=exchange_token, token_type='eg1')
        elif authorization_code:
            params = dict(grant_type='authorization_code', code=authorization_code, token_type='eg1')
        elif client_credentials:
            params = dict(grant_type='client_credentials', token_type='eg1')
        else:
            raise ValueError('At least one token type must be specified!')

        url = f'{self._oauth_host}/account/api/oauth/token'
        r = self.session.post(url, data=params, auth=self._oauth_basic, timeout=self.timeout)
        # Only raise HTTP exceptions on server errors
        if r.status_code >= 500:
            r.raise_for_status()

        j = r.json()
        if 'errorCode' in j:
            if j['errorCode'] == 'errors.com.epicgames.oauth.corrective_action_required':
                self.logger.error(
                    f'{j["errorMessage"]} ({j["correctiveAction"]}), '
                    f'open the following URL to take action: {j["continuationUrl"]}'
                )
            else:
                self.logger.error(f'Login to EGS API failed with errorCode: {j["errorCode"]}')
            raise InvalidCredentialsError(j['errorCode'])
        elif r.status_code >= 400:
            self.logger.error(f'EGS API responded with status {r.status_code} but no error in response: {j}')
            raise InvalidCredentialsError('Unknown error')

        self.session.headers['Authorization'] = f'bearer {j["access_token"]}'

        # only set user info when using non-anonymous login
        if not client_credentials:
            self.user = j

        if self.debug_mode:
            self.save_auth_data(refresh_token, exchange_token, authorization_code)

        return j

    def get_item_token(self) -> str:
        """
        Get the item token.
        Unused but kept for the global API reference.
        :return: item token using json format.
        """
        url = f'{self._oauth_host}/account/api/oauth/exchange'
        r = self.session.get(url, timeout=self.timeout)
        r.raise_for_status()
        return r.json()

    def get_item_manifest(self, namespace, catalog_item_id, app_name, platform='Windows', label='Live') -> dict:
        """
        Get the item manifest.
        :param namespace:  namespace.
        :param catalog_item_id: catalog item id.
        :param app_name: Asset name.
        :param platform: platform to get manifest for.
        :param label: label of the manifest.
        :return: item manifest using json format.
        """
        url = f'{self._launcher_host}/launcher/api/public/assets/v2/platform/{platform}/namespace/{namespace}/catalogItem/{catalog_item_id}/app/{app_name}/label/{label}'
        r = self.session.get(url, timeout=self.timeout)
        r.raise_for_status()
        return r.json()

    def get_item_info(self, namespace: str, catalog_item_id: str) -> (dict, int):
        """
        Get the item info.
        :param namespace: namespace of the item.
        :param catalog_item_id: catalog item id of the item.
        :return: (The item info, status code).
        """
        url = f'https://{self._catalog_host}/catalog/api/shared/namespace/{namespace}/bulk/items'
        r = self.session.get(
            url,
            params=dict(
                id=catalog_item_id, includeDLCDetails=True, includeMainGameDetails=True, country=self.country_code, locale=self.language_code
            ),
            timeout=self.timeout
        )
        r.raise_for_status()
        return r.json().get(catalog_item_id, None), r.status_code

    def get_asset_data_from_marketplace(self, url: str) -> dict:
        """
        Get the asset data from the marketplace using beautifulsoup.
        :param url: url to grab.

        Notes:
            This is the only way I know to get the id of an asset from its slug (or url)
        """
        empty_data = {
            'id': '',
            'name': '',
            'category': '',
            'description': '',
            'image': '',
            'release_date': '',
            'url': '',
            'price': '',
            'price_currency': '',
            'page_title': '',
            'grab_result': '',
        }
        json_data = empty_data.copy()
        asset_slug = url.split('/')[-1]
        json_data['page_title'] = asset_slug
        response = self.get_url_with_uc(url)
        if response.ok and response.content:
            self.logger.info(f'Grabbing asset data from {url}')
        else:
            self.logger.warning(f'Can not get asset data for {url}')
            json_data['grab_result'] = GrabResult.PAGE_NOT_FOUND.name
            return json_data

        json_data['grab_result'] = GrabResult.NO_ERROR.name

        soup = BeautifulSoup(response.content, 'html.parser')

        # Finding all script tags in the HTML
        # scripts = soup.find_all('script')
        # Finding all script tags with the type 'application/ld+json' in the HTML
        scripts = soup.find_all('script', {'type': 'application/ld+json'})

        for index, script in enumerate(scripts):
            # Remove script tags
            json_text = script.string

            # Convert to Python dictionary
            json_dict = json.loads(json_text)

            # check if the script describes a product
            if json_dict['@type'] == 'Product':
                try:
                    json_data['id'] = json_dict['productID']
                    json_data['name'] = json_dict['name']
                    json_data['category'] = json_dict['category']
                    json_data['description'] = json_dict['description']
                    json_data['image'] = json_dict['image']
                    json_data['release_date'] = json_dict['releaseDate']
                    json_data['url'] = json_dict['offers']['url']  # must be the same as url
                    json_data['price'] = json_dict['offers']['price']
                    json_data['price_currency'] = json_dict['offers']['priceCurrency']
                except KeyError as error:
                    self.logger.warning(f"A key is missing in Script {index + 1}: {error!r}")
                    json_data['grab_result'] = GrabResult.PARTIAL.name
                    continue
        if json_data['url'].lower().replace('www.', '') != url.lower().replace('www.', ''):
            self.logger.warning(f"URLs do not match: {json_data['url']} != {url}")
            json_data['grab_result'] = GrabResult.INCONSISTANT_DATA.name
        return json_data

    def save_auth_data(self, refresh_token: str = None, exchange_token: str = None, authorization_code: str = None) -> None:
        """
        Save user and authorization data into files. This is useful for debugging.
        :param refresh_token:
        :param exchange_token:
        :param authorization_code:
        """
        keep_existing = True  # True is usefull to keep the authorization_code created at first loggin
        auth_folder = Path('../.private/auth_data')  # TODO: add the path to settings ?
        # delete the content folder if it exists
        if auth_folder.exists():
            if not keep_existing:
                for file in auth_folder.iterdir():
                    file.unlink()
        else:
            auth_folder.mkdir(parents=True, exist_ok=True)
        # COOKIES: see: https://scrapfly.io/blog/save-and-load-cookies-in-requests-python/
        cookies = dict_from_cookiejar(self.session.cookies)
        if cookies:
            Path(f'{auth_folder}/session_cookies.json').write_text(json.dumps(cookies))
        # HEADERS:
        # NOTE! self.session.headers is a CaseInsensitiveDict object, not a dict
        headers = dict(self.session.headers)
        if headers:
            Path(f'{auth_folder}/session_headers.json').write_text(json.dumps(headers))
        # Tokens
        if refresh_token:
            Path(f'{auth_folder}/refresh_token.token').write_text(refresh_token)
        if exchange_token:
            Path(f'{auth_folder}/exchange_code.token').write_text(exchange_token)
        if authorization_code:
            Path(f'{auth_folder}/authorization_code.token').write_text(authorization_code)

    def get_url_with_uc(self, url: str, timeout=None, force_bypass_captcha: bool = False) -> Response:
        """
        Return the response of an url request:
        - using the Nodriver object if the url concerns the epic marketplace.
        - using the session object if not
        :param url: url to fix.
        :param timeout: timeout for the request.
        :param force_bypass_captcha: force the bypass of the captcha.
        :return: response Object.

        NOTES:
            When using Nodriver,
            The parameters for the browser and the request are stored in self._uc_request.params
            The response will be stored in self._uc_request.response.
        """
        if not timeout:
            timeout = self.timeout

        uc_response = UCResponse()
        uc_response.request = self.session.request
        # get the connection object from the current session
        # we use self._url_marketplace as default url for the connection
        adapter = cast(HTTPAdapter, self.session.get_adapter(self._url_marketplace))
        connection = adapter.get_connection(self._url_marketplace)
        uc_response.connection = connection
        self._uc_request.request_type = UCRequestType.NORMAL  # by default, we consider we are using a normal request
        self._uc_request.response = uc_response

        # method_for_get_session = 0 # use a "normal" session get. Will not bypass a captcha if present
        # method_for_get_session = 1 # nodriver
        # method_for_get_session = 2 # reqdriver

        if force_bypass_captcha or self.bypass_captcha or url.startswith(self._url_asset_list) or url.startswith(self._url_search_asset):
            method_for_get_session = 1
        else:
            method_for_get_session = 0  # simple session get

        # force a method
        # method_for_get_session = 0  # debug only

        if method_for_get_session == 1:
            # test nodriver
            # install by: pip install nodriver
            # RESULT:
            #   response is OK, BUT the user is not logged in

            del self.session.cookies['EPIC_CLIENT_SESSION']  # this data cause a http 400 error because its value is too big
            # noinspection GrazieInspection
            # has_captcha = response.content.find(b"Please complete a security check to continue") != -1
            if not self._uc_browser:
                self.init_uc_browser()
            uc.loop().run_until_complete(self.uc_get_content(url))
            # note that self._uc_request is updated by the uc_get_content method
            return self._uc_request.response
        elif method_for_get_session == 2:
            # test reqdriver
            # install by: pip install reqdriver

            # RESULT:
            #   HS: response is empty
            # noinspection PyUnresolvedReferences
            from reqdriver import RequestsDriver
            from selenium.webdriver.chrome.options import Options
            custom_options = Options()
            custom_options.add_argument('--headless')
            if not self._uc_browser:
                req_driver = RequestsDriver(self.session, custom_options=custom_options)
                self._uc_browser = req_driver.get_driver()
                result_dict: dict = self._uc_browser.execute(Command.GET, {"url": url})
                uc_response.content = result_dict['value']
                uc_response.status_code = 200 if uc_response.content else 403
                self._uc_request.response = uc_response
                self._uc_request.request_type = UCRequestType.USING_UD
                return uc_response
        else:
            # use a "normal" session get. Will not bypass the captcha if present
            return self.session.get(url, timeout=timeout)

    def init_uc_browser(self) -> None:
        """
        Initialize the undetected Chrome Browser.

        NOTES:
            The parameters for the browser and the request are stored in self._uc_request.params
            The response will be stored in self._uc_request.response.
        """
        window_width = 1024 if self.debug_mode else 100
        window_height = 768 if self.debug_mode else 100
        # note:  the window position could be displayed OUTSIDE the current screen, and it works !!
        window_left = -window_width if not self.debug_mode else 0
        window_top = -window_height if not self.debug_mode else 0

        params = {}
        browser_path = r'C:\Program Files\Chromium\Application\chrome.exe'  # TODO: add the default browser_path to the settings or add a new check to find the browser
        params['headless'] = False  # True  will not bypass the captcha
        params['use_subprocess'] = True  # False will crash the script
        params['browser_executable_path'] = browser_path
        params['user_multi_procs'] = True  # will speed up the process

        # set some arguments for the browser
        # full list: https://peter.sh/experiments/chromium-command-line-switches/
        params['browser_args'] = [
            f'--window-size={window_width},{window_height}',  # set the initial size of the window
            f'--window-position={window_left},{window_top}',  # set the initial position of the window
            '--window-name="UndetectedChrome"',  # set the name of the window
            '--no-first-run', '--no-default-browser-check', '--no-experiments', '--mute-audio', '--enable-gpu', '--disable-extensions',
            # next options have been tested and are OK
            # '--start-maximized', # debug only
            # '--incognito',

            # next options have been tested and could create issue
            # '--no-sandbox',  # needed for linux

            # next options have been tested and CAUSE AN ERROR
            # '--headless',
            # '--silent-launch',
        ]
        self._uc_request.params = params

        response = self._uc_request.response
        response.request = self.session.request
        # get the connection object from the current session
        # we use self._url_marketplace as default url for the connection
        adapter = cast(HTTPAdapter, self.session.get_adapter(self._url_marketplace))
        connection = adapter.get_connection(self._url_marketplace)
        response.connection = connection
        self._uc_request.response = response
        uc_logger = logging.getLogger('nodriver.core.browser')
        uc_logger.setLevel(logging.WARNING)  # disable info level because too verbose

    async def uc_get_content(self, url) -> None:
        """
        Get the content of a page using the undetected Chrome driver.
        :param url: url to get the content from.

        NOTES:
            The response will be stored in self._uc_request.response.
        """
        self._uc_browser = await uc.start(**self._uc_request.params)
        page = await self._uc_browser.get(url)
        await page.activate()
        raw_content = await page.get_content()
        await page.close()
        response = self._uc_request.response
        response.headers = self.session.headers
        response.raw = raw_content
        response.url = url
        response.status_code = 200 if raw_content else 403
        self._uc_request.request_type = UCRequestType.USING_UD
        self._uc_request.response = response
        if response.status_code == 403:
            return

        soup = BeautifulSoup(raw_content, 'html.parser')
        if soup:
            # get the charset value from the header of raw_content
            charset_tag = soup.find('meta', charset=True)
            response.encoding = charset_tag.get('charset', response.encoding)

            # if response contains json data, raw_content is a container where the real content is stored between <pre> and </pre>
            pre_tag = soup.find('pre')
            response.content = pre_tag.text if pre_tag else raw_content

            # convert content to bytes because it's needed for the json decoding in response.json()
            response.content = response.content.encode(response.encoding)
            response.status_code = 200
        self._uc_request.response = response

    def get_last_request_type(self) -> UCRequestType:
        """
        Return the type of the last request.
        """
        return self._uc_request.request_type



# coding=utf-8
"""
Definition for the types used in this module:
- GrabResult: enum for the result of grabbing a page
- UEAssetType: enum to represent the asset type
- DataSourceType: enum to represent the data source type
- WidgetType: enum for the widget types.
"""
from enum import Enum


class UCRequestType(Enum):
    """
    Enum for the type of request.
    """
    NORMAL = 1,
    USING_UD = 2


class GrabResult(Enum):
    """
    Enum for the result of grabbing a page.
    """
    # Keep this class here (and not in egs.py for instance) to limit circular import.
    NO_ERROR = 0
    INCONSISTANT_DATA = 1
    PAGE_NOT_FOUND = 2
    CONTENT_NOT_FOUND = 3
    TIMEOUT = 4
    PARTIAL = 5  # when asset has been added when owned asset data only (less complete that "standard" asset data)
    NO_APPID = 6  # no appid found in the data (will produce a file name like '_no_appId_asset_1e10acc0cca34d5c8ff7f0ab57e7f89f
    NO_RESPONSE = 7  # the url does not return HTTP 200


class DataFrameUsed(Enum):
    """ Enum to represent the data frame used for getting or setting the data"""
    AUTO = 0  # Automatically select the data frame wether it is filtered or not
    UNFILTERED = 1  # Use the data frame with unfiltered data
    FILTERED = 2  # Use the data frame with filtered data
    MODEL = 3  # Use the model.df.
    BOTH = 4  # Use both data frames


class UEAssetType(Enum):
    """ Enum to represent the asset type """
    Unknown = 0
    Plugin = 1
    Asset = 2
    Manifest = 3

    @property
    def category_name(self):
        """ Return the category name of the asset type """
        if self == self.Plugin:
            return 'local/Plugin'  # existing category in the marketplace
        if self == self.Asset:
            return 'local/Asset'  # non-existing category in the marketplace
        if self == self.Manifest:
            return 'local/Manifest'  # non-existing category in the marketplace
        return 'local/unknown'  # non-existing category in the marketplace


class DataSourceType(Enum):
    """ Enum to represent the data source type """
    FILE = 1
    DATABASE = 2


class WidgetType(Enum):
    """ Enum for the widget types """
    ENTRY = 0  # Entry widget
    TEXT = 1  # Text widget
    LABEL = 2  # Label widget
    CHECKBUTTON = 3  # Checkbutton widget
    BUTTON = 4  # Button widget


class FilterType(Enum):
    """ Enum for the filter types """
    STR = 0  # filter use a string as query
    CALLABLE = 1  # filter use a callable (from FilterCallableClass)
    LIST = 2  # filter use a list of value for an asset_id or app_name (json encoded)

    @classmethod
    def from_name(cls, name: str) -> 'FilterType':
        """ Return the filter type from its name """
        for ftype in cls:
            if ftype.name.lower() == name.lower():
                return ftype
        return cls.STR



# coding: utf-8
"""
implementation for:
- EPCLFS: Epic Games Encrypted Config Filesystem.
"""
import configparser
import os

from UEVaultManager.lfs.utils import path_join


class EPCLFS:
    """
    Epic Games Encrypted Config Filesystem Class.
    """
    data_keys = []

    def __init__(self):
        self.manifests = {}
        self.is_windows = os.name == 'nt'
        if self.is_windows:
            self.userprofile_path = os.path.normpath(os.path.expandvars(r'$USERPROFILE'))
            self.appdata_path = os.path.normpath(os.path.expandvars(r'%LOCALAPPDATA%\EpicGamesLauncher\Saved\Config\Windows'))
            self.programdata_path = os.path.normpath(os.path.expandvars(r'%PROGRAMDATA%\Epic\EpicGamesLauncher'))
        else:
            self.userprofile_path = os.path.normpath(os.path.expandvars(r'$HOME'))
            self.appdata_path = self.programdata_path = None
        self.config = configparser.ConfigParser(strict=False)
        self.vault_cache_folder = ''
        self.projects_path = ''
        self.engines_path = ''
        self.read_config()

    def read_config(self):
        """
        Read the EGS config files.
        """
        if not self.appdata_path:
            raise ValueError('EGS AppData path is not set')

        if not os.path.isdir(self.appdata_path):
            raise ValueError('EGS AppData path does not exist')

        self.config.read(path_join(self.appdata_path, 'GameUserSettings.ini'), encoding='utf-8')
        fallback = path_join(self.programdata_path, 'VaultCache') if self.is_windows else '/opt/Epic Games/VaultCache'
        self.vault_cache_folder = self.config.get('Launcher', 'VaultCacheDirectories', fallback=fallback)
        fallback = path_join(self.userprofile_path, '/Documents/Unreal Projects/')
        self.projects_path = self.config.get('Launcher', 'CreatedProjectPaths', fallback=fallback)
        fallback = 'C:/Program Files/Epic Games' if self.is_windows else '/opt/Epic Games'
        self.engines_path = self.config.get('Launcher', 'DefaultAppInstallLocation', fallback=fallback)



# coding: utf-8
"""
implementation for:
- AssetBase: Asset base data
- Asset: Combination of Asset, Asset metadata and Asset extra as stored on disk
"""
from dataclasses import dataclass, field
from typing import Dict, List


@dataclass
class AssetBase:
    """
    An asset with minimal data.
    """
    app_name: str = ''
    asset_id: str = ''
    build_version: str = ''
    catalog_item_id: str = ''
    label_name: str = ''
    namespace: str = ''
    metadata: Dict = field(default_factory=dict)

    # noinspection DuplicatedCode
    @classmethod
    def from_egs_json(cls, json) -> 'AssetBase':
        """
        Create AssetBase from EGS.
        :param json: data.
        :return: AssetBase object.
        """
        tmp = cls()
        tmp.app_name = json.get('appName', '')
        tmp.asset_id = json.get('assetId', '')
        tmp.build_version = json.get('buildVersion', '')
        tmp.catalog_item_id = json.get('catalogItemId', '')
        tmp.label_name = json.get('labelName', '')
        tmp.namespace = json.get('namespace', '')
        tmp.metadata = json.get('metadata', {})
        return tmp

    # noinspection DuplicatedCode
    @classmethod
    def from_json(cls, json) -> 'AssetBase':
        """
        Create AssetBase from json.
        :param json: data.
        :return: AssetBase object.
        """
        tmp = cls()
        tmp.app_name = json.get('app_name', '')
        tmp.asset_id = json.get('asset_id', '')
        tmp.build_version = json.get('build_version', '')
        tmp.catalog_item_id = json.get('catalog_item_id', '')
        tmp.label_name = json.get('label_name', '')
        tmp.namespace = json.get('namespace', '')
        tmp.metadata = json.get('metadata', {})
        return tmp


@dataclass
class Asset:
    """
    Combination of a base asset, asset metadata and asset extra as stored on disk.
    """
    app_name: str
    app_title: str
    asset_infos: Dict[str, AssetBase] = field(default_factory=dict)
    base_urls: List[str] = field(default_factory=list)
    metadata: Dict = field(default_factory=dict)
    size: str = ''  # only used when asset when scanning for asset in File Mode. Will only be set when the cli.list_assets() method is called.

    def app_version(self, platform: str = 'Windows'):
        """
        Get Asset version for a given platform.
        :param platform: platform.
        :return: asset version.
        """
        if platform not in self.asset_infos:
            return None
        return self.asset_infos[platform].build_version

    @property
    def catalog_item_id(self):
        """
        Get catalog item id.
        :return: catalog item id.
        """
        if not self.metadata:
            return None
        return self.metadata['id']

    @property
    def namespace(self):
        """
        Get namespace.
        :return: namespace.
        """
        if not self.metadata:
            return None
        return self.metadata['namespace']

    @classmethod
    def from_json(cls, asset_data: dict) -> 'Asset':
        """
        Create Asset from json.
        :param asset_data: data.
        :return: Asset object.
        """
        tmp = cls(app_name=asset_data.get('app_name', ''), app_title=asset_data.get('app_title', ''), )  # call to the class constructor
        tmp.metadata = asset_data.get('metadata', {})
        if 'asset_infos' in asset_data:
            try:
                tmp.asset_infos = {k: AssetBase.from_json(v) for k, v in asset_data['asset_infos'].items()}
            except AttributeError:
                tmp.asset_infos = {}
        else:
            # Migrate old asset_info to new asset_infos
            tmp.asset_infos['Windows'] = AssetBase.from_json(asset_data.get('asset_info', {}))
        tmp.base_urls = asset_data.get('base_urls', [])
        return tmp

    @property
    def __dict__(self):
        """This is just here so asset_infos gets turned into a dict as well"""
        try:
            assets_dict = {k: v.__dict__ for k, v in self.asset_infos.items()}
        except AttributeError:
            assets_dict = {}
        return dict(metadata=self.metadata, asset_infos=assets_dict, app_name=self.app_name, app_title=self.app_title, base_urls=self.base_urls)


@dataclass
class InstalledAsset:
    """
    Local metadata for an installed asset
    """
    app_name: str
    catalog_item_id: str = ''
    title: str = ''
    version: str = ''
    base_urls: List[str] = field(default_factory=list)
    egl_guid: str = ''
    install_size: int = 0
    manifest_path: str = ''
    platform: str = 'Windows'
    installed_folders: List[str] = field(default_factory=list)

    @classmethod
    def from_json(cls, asset_data: dict) -> 'InstalledAsset':
        """
        Create InstalledAsset from json.
        :param asset_data: data.
        :return: InstalledAsset object.
        """
        tmp = cls(
            app_name=asset_data.get('app_name', ''),
            catalog_item_id=asset_data.get('catalog_item_id', ''),
            installed_folders=asset_data.get('installed_folders', []),
            title=asset_data.get('title', ''),
            version=asset_data.get('version', ''),
        )
        tmp.base_urls = asset_data.get('base_urls', [])
        tmp.egl_guid = asset_data.get('egl_guid', '')
        tmp.install_size = asset_data.get('install_size', 0)
        tmp.manifest_path = asset_data.get('manifest_path', '')
        tmp.platform = asset_data.get('platform', 'Windows')
        return tmp

    @property
    def install_path(self) -> str:
        """
        Get the "install path" of the installed asset.
        :return: install path.

        Notes:
            Will return the most recent values from installed_folders property.
        """
        install_path = ''
        if isinstance(self.installed_folders, list) and len(self.installed_folders) > 0:
            install_path = self.installed_folders[-1]
        return install_path

    @install_path.setter
    def install_path(self, path: str):
        """
        Set an "install path"
        :param path: install path.

        Notes:
            Add the path at the end of the installed_folders proper ty.
        """
        path = path.strip()
        if not path or path in self.installed_folders:
            return
        self.installed_folders.append(path)
